"""
Real picture based test sequences.
"""

from vc2_conformance.test_cases import encoder_test_case_generator

from vc2_conformance.test_cases.encoder.common import picture_generator_to_test_case

from vc2_conformance import picture_generators


@encoder_test_case_generator
def real_pictures(codec_features):
    """
    **Tests real pictures are encoded sensibly.**

    This test contains a series of three still photographs of natural scenes
    with varying levels of high and low spatial frequency content.

    .. image:: /_static/user_guide/real_pictures.svg

    .. note::

        The source images for this test case are generated from 4256 by 2832
        pixel, 4:4:4, 16 bit, standard dynamic range, RGB color images with the
        ITU-R BT.709 gamut. As such, the pictures might be of reduced technical
        quality compared with the capabilities of the format. The rescaling and
        color conversion algorithms used are also basic in nature, potentially
        further reducing the picture quality.
    """
    return picture_generator_to_test_case(
        picture_generators.real_pictures,
        codec_features,
    )



r"""
.. _vc2-bitstream-validator:

``vc2-bitstream-validator``
===========================

A command-line utility for validating VC-2 bitstreams' conformance with the
VC-2 specification and providing a reference decoding of the pictures within.

Usage
-----

This command should be passed a filename containing a candidate VC-2 bitstream.
For example, given a valid stream::

    $ vc2-bitstream-validator path/to/bitstream.vc2 --output decoded_picture_%d.raw
    No errors found in bitstream. Verify decoded pictures to confirm conformance.

Here the ``--output`` argument specifies the printf-style template for the
decoded picture filenames. The decoded pictures are written as raw files (see
:ref:`file-format`).

If a conformance error is detected, a detailed explanation of the problem is
displayed::

    $ vc2-bitstream-validator invalid.vc2
    Conformance error at bit offset 104
    ===================================

    Non-zero previous_parse_offset, 5789784, in the parse info at the start of
    a sequence (10.5.1).


    Details
    -------

    Was this parse info block copied from another stream without the
    previous_parse_offset being updated?

    Does this parse info block incorrectly include an offset into an adjacent
    sequence?


    Suggested bitstream viewer commands
    -----------------------------------

    To view the offending part of the bitstream:

        vc2-bitstream-viewer invalid.vc2 --offset 104


    Pseudocode traceback
    --------------------

    Most recent call last:

    * parse_stream (10.3)
      * parse_sequence (10.4.1)
        * parse_info (10.5.1)

    vc2-bitstream-validator: error: non-conformant bitstream (see above)

Errors include an explanation of the conformance problem (along with references
to the VC-2 standards documents) along with possible causes of the error.
Additionally, a sample invocation of :ref:`vc2-bitstream-viewer` is given which
can be used to display the contents of the bitstream at the offending position.
Finally, a stack trace for the VC-2 pseudocode functions involved in parsing
the stream at the point of failure is also printed.


Arguments
---------

The complete set of arguments can be listed using ``--help``

.. program-output:: vc2-bitstream-validator --help

"""

import os
import sys
import traceback

from argparse import ArgumentParser

from textwrap import dedent

from vc2_conformance.pseudocode.metadata import make_pseudocode_traceback

from vc2_conformance import __version__

from vc2_conformance.py2x_compat import quote

from vc2_conformance.string_utils import wrap_paragraphs

from vc2_conformance.file_format import write

from vc2_conformance.pseudocode.state import State

from vc2_conformance.decoder import (
    init_io,
    parse_stream,
    ConformanceError,
    tell,
)

from vc2_conformance.bitstream import to_bit_offset

from vc2_conformance.py2x_compat import get_terminal_size


def format_pseudocode_traceback(tb):
    """
    Given a :py:func:`traceback.extract_tb` generated traceback description,
    return a string describing the current stack of VC-2 pseudocode functions
    being called.
    """
    ptb = make_pseudocode_traceback(tb)

    return "\n".join(
        "{}* {}".format("  " * num, pdf.citation) for num, pdf in enumerate(ptb)
    )


class BitstreamValidator(object):
    def __init__(self, filename, show_status, verbose, output_filename):
        """
        Parameters
        ==========
        filename : str
            The bitstream filename to read from.
        show_status : bool
            If True, show a status line indicating progress during validation.
        verbose : int
            If >=1, show Python stack traces on failure.
        output_filename : str
            A filename pattern for output bitstream files. Should contain a
            printf-style format string (e.g. "picture_%d.raw").
        """
        self._filename = filename
        self._show_status = show_status
        self._verbose = verbose
        self._output_filename = output_filename

        # The index to use in the filename of the next decoded picture
        self._next_picture_index = 0

        # Is the status line currently visible
        self._status_line_visible = False

    def run(self):
        try:
            self._file = open(self._filename, "rb")
            self._filesize_bytes = os.path.getsize(self._filename)
        except Exception as e:
            # Catch-all exception handler excuse: Catching only file-related
            # exceptions is challenging, particularly in a backward-compatible
            # manner. However, none of the above are known to produce
            # exceptions *except* due to file-related issues.
            self._print_error(str(e))
            return 1

        self._state = State(_output_picture_callback=self._output_picture)
        init_io(self._state, self._file)

        if self._show_status:
            self._update_status_line("Starting bitstream validation...")

        try:
            parse_stream(self._state)
            self._hide_status_line()
            if tell(self._state) == (0, 7):
                sys.stdout.flush()
                sys.stderr.write("Warning: 0 bytes read, bitstream is empty.\n")
            print(
                "No errors found in bitstream. Verify decoded pictures to confirm conformance."
            )
            return 0
        except ConformanceError as e:
            # Bitstream failed validation
            exc_type, exc_value, exc_tb = sys.exc_info()
            self._hide_status_line()
            self._print_conformance_error(e, traceback.extract_tb(exc_tb))
            self._print_error("non-conformant bitstream (see above)")
            return 2
        except Exception as e:
            # Internal error (shouldn't happen(!))
            self._hide_status_line()
            self._print_error(
                "internal error in bitstream validator: {}: {} "
                "(probably a bug in this program)".format(
                    type(e).__name__,
                    str(e),
                )
            )
            return 3

    def _output_picture(self, picture, video_parameters, picture_coding_mode):
        filename = self._output_filename % (self._next_picture_index,)
        self._next_picture_index += 1

        write(
            picture,
            video_parameters,
            picture_coding_mode,
            filename,
        )

        if self._show_status:
            self._update_status_line("Decoded picture written to {}".format(filename))

    def _update_status_line(self, message):
        """
        Display/update the status line indicating the progress of the decoding
        process.
        """
        self._status_line_visible = True

        percent = int(
            round((tell(self._state)[0] * 100.0) / (self._filesize_bytes or 1))
        )

        line = "[{:3d}%] {}".format(percent, message)

        # Ensure stdout is fully displayed before doing anything to the status
        # line.
        sys.stdout.flush()

        sys.stderr.write(
            (
                "\033[2K"  # Clear to end of line
                "\033[s"  # Save cursor position
                "{}"
                "\033[u"  # Restore cursor position
            ).format(line)
        )
        sys.stderr.flush()

    def _hide_status_line(self):
        """If the status line is visible, hide it."""
        if self._status_line_visible:
            self._status_line_visible = False

            # Ensure stdout is fully displayed before doing anything to the
            # status line.
            sys.stdout.flush()

            sys.stderr.write("\033[2K")  # Clear to end of line
            sys.stderr.flush()

    def _print_conformance_error(self, exception, tb):
        """
        Display detailed information from a ConformanceError on stdout.
        """
        terminal_width = get_terminal_size()[0]

        summary, _, details = wrap_paragraphs(exception.explain()).partition("\n")

        offending_offset = exception.offending_offset()
        if offending_offset is None:
            offending_offset = to_bit_offset(*tell(self._state))

        title = "Conformance error at bit offset {}".format(offending_offset)

        bitstream_viewer_hint = (
            dedent(exception.bitstream_viewer_hint())
            .strip()
            .format(
                cmd="vc2-bitstream-viewer",
                file=quote(self._filename),
                offset=offending_offset,
            )
        )

        out = ""

        out += title + "\n"
        out += ("=" * len(title)) + "\n"
        out += "\n"
        out += wrap_paragraphs(summary, terminal_width) + "\n"
        out += "\n"
        out += "\n"
        out += "Details\n"
        out += "-------\n"
        out += "\n"
        out += wrap_paragraphs(details, terminal_width) + "\n"
        out += "\n"
        out += "\n"
        out += "Suggested bitstream viewer commands\n"
        out += "-----------------------------------\n"
        out += "\n"
        out += bitstream_viewer_hint + "\n"
        out += "\n"
        out += "\n"
        out += "Pseudocode traceback\n"
        out += "--------------------\n"
        out += "\n"
        out += "Most recent call last:\n"
        out += "\n"
        out += format_pseudocode_traceback(tb) + "\n"

        print(out)

    def _print_error(self, message):
        """
        Print an error message to stderr.
        """
        # Avoid interleaving with stdout (and make causality clearer)
        sys.stdout.flush()

        # Display the traceback
        if self._verbose >= 1:
            if sys.exc_info()[0] is not None:
                traceback.print_exc()

        # Display the message
        prog = os.path.basename(sys.argv[0])
        message = "{}: error: {}".format(prog, message)
        sys.stderr.write("{}\n".format(message))


def parse_args(*args, **kwargs):
    """
    Parse a set of command line arguments. Returns a :py:mod:`argparse`
    ``args`` object with the following fields:

    * bitstream (str): The filename of the bitstream to read
    * no_status (bool): True if the status line is to be hidden.
    * verbose (int): The verbosity level.
    * output (str): The output picture filename pattern.
    """
    parser = ArgumentParser(
        description="""
        Validate a bitstream's conformance with the VC-2 specifications.
    """
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__),
    )

    parser.add_argument(
        "bitstream",
        help="""
            The filename of the bitstream to validate.
        """,
    )

    parser.add_argument(
        "--no-status",
        "--quiet",
        "-q",
        action="store_true",
        default=False,
        help="""
            Do not display a status line on stderr while validating the
            bitstream.
        """,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="""
            Show full Python stack-traces on failure.
        """,
    )

    parser.add_argument(
        "--output",
        "-o",
        default="picture_%d.raw",
        help="""
            The filename pattern for decoded picture data and metadata. The
            supplied pattern should a 'printf' style template with (e.g.) '%%d'
            where an index will be substituted. The first decoded picture will
            be assigned index '0', the second '1' and so on -- i.e. these
            indices are unrelated to the picture number. The file extension
            supplied will be stripped and two files will be written for each
            decoded picture: a '.raw' planar image file and a '.json' JSON
            metadata file. (Default: %(default)s).
        """,
    )

    args = parser.parse_args(*args, **kwargs)

    try:
        args.output % (0,)
    except TypeError as e:
        parser.error("--output is not a valid printf template: {}".format(e))

    return args


def main(*args, **kwargs):
    args = parse_args(*args, **kwargs)

    validator = BitstreamValidator(
        filename=args.bitstream,
        show_status=not args.no_status,
        verbose=args.verbose,
        output_filename=args.output,
    )
    return validator.run()


if __name__ == "__main__":
    sys.exit(main())



"""
The :py:mod:`vc2_conformance.bitstream.metadata` module contains metadata about
the relationship between VC-2 pseudocode functions and deserialised
:py:mod:`~vc2_conformance.fixeddict` structures. These are used by the
:ref:`vc2-bitstream-viewer` command to produce richer output and during
documentation generation.

.. autodata:: pseudocode_function_to_fixeddicts
    :annotation:

.. autodata:: pseudocode_function_to_fixeddicts_recursive
    :annotation:

.. autodata:: fixeddict_to_pseudocode_function
    :annotation:

"""

from vc2_conformance.bitstream import vc2
from vc2_conformance.bitstream.vc2_fixeddicts import (
    vc2_fixeddict_nesting,
    HQSlice,
    LDSlice,
)


__all__ = [
    "pseudocode_function_to_fixeddicts",
    "pseudocode_function_to_fixeddicts_recursive",
    "fixeddict_to_pseudocode_function",
]


pseudocode_function_to_fixeddicts = {}
"""
For the subset of pseudocode functions in the VC-2 spec dedicated to
serialisation/deserialisation, gives the corresponding fixeddict type in
:py:mod:`vc2_conformance.bitstream.vc2_fixeddicts`.

A dictionary of the shape ``{function_name: [type, ...], ...}``.
"""

# Populate pseudocode_function_to_fixeddicts
for name in vc2.__all__:
    function = getattr(vc2, name)
    if hasattr(function, "context_type"):
        pseudocode_function_to_fixeddicts[name] = [function.context_type]
    else:
        pseudocode_function_to_fixeddicts[name] = []

# Special case: 'slice' may have one of two types and this is not indicated by
# a @context_type decorator
assert "slice" in pseudocode_function_to_fixeddicts
assert pseudocode_function_to_fixeddicts["slice"] == []
pseudocode_function_to_fixeddicts["slice"].append(HQSlice)
pseudocode_function_to_fixeddicts["slice"].append(LDSlice)

# Special case: 'slice_band' and 'color_diff_slice_band' both only serve to
# update an existing slice and so don't have a type. Remove these from the
# listing to avoid being misleading
assert pseudocode_function_to_fixeddicts.pop("slice_band") == []
assert pseudocode_function_to_fixeddicts.pop("color_diff_slice_band") == []


fixeddict_to_pseudocode_function = {
    fixeddicts[0]: function_name
    for function_name, fixeddicts in pseudocode_function_to_fixeddicts.items()
    # Special case: prefer the type-specific variants of 'slice'
    if function_name != "slice"
}
"""
Provides a mapping from :py:mod:`vc2_conformance.bitstream.vc2_fixeddicts` types to the name of the
corresponding pseudocode function which may be used to serialise/deserialise
from/to it.
"""


pseudocode_function_to_fixeddicts_recursive = {}
"""
Like :py:data:`pseudocode_function_to_fixeddicts` but each entry also
recursively includes the fixeddict types of all contained entries.
"""


# Populate pseudocode_function_to_fixeddicts_recursive
def iter_fixeddict_types(fixeddict):
    yield fixeddict

    for nested_type in vc2_fixeddict_nesting.get(fixeddict, []):
        for subtype in iter_fixeddict_types(nested_type):
            yield subtype


for name, base_types in pseudocode_function_to_fixeddicts.items():
    types = []
    for base_type in base_types:
        types.extend(iter_fixeddict_types(base_type))
    pseudocode_function_to_fixeddicts_recursive[name] = types



import os

from setuptools import setup, find_packages

version_file = os.path.join(os.path.dirname(__file__), "vc2_conformance", "version.py")
with open(version_file, "r") as f:
    exec (f.read())  # noqa: E211

readme_file = os.path.join(os.path.dirname(__file__), "README.md")
with open(readme_file, "r") as f:
    long_description = f.read()

setup(
    name="vc2_conformance",
    version=__version__,  # noqa: F821 -- loaded by 'exec' above
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/bbc/vc2_conformance",
    author="BBC R&D",
    description="Conformance testing utilities for the VC-2 video codec.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPL-3.0-only",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Telecommunications Industry",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    keywords="vc2 dirac dirac-pro conformance",
    install_requires=[
        # NB: bitarray-hardbyte is used in place of official 'bitarray' package
        # due to provision of a binary wheel.
        "bitarray-hardbyte",
        "sentinels",
        "vc2_data_tables >=0.1.1, <2.0",
        "vc2_bit_widths >=0.1.1, <2.0",
        "vc2_conformance_data >=0.1.1, <2.0",
        # Use old versions/polyfill libraries which have deprecated older Python
        # version support
        "enum34; python_version<'3.4'",
        "pillow<7; python_version<'3.0'",
        "pillow; python_version>='3.0'",
        "numpy<1.17; python_version<'3.0'",
        "numpy<1.20; python_version>='3.0' and python_version<'3.7'",
        "numpy; python_version>='3.7'",
    ],
    entry_points={
        "console_scripts": [
            "vc2-bitstream-viewer=vc2_conformance.scripts.vc2_bitstream_viewer:main",
            "vc2-bitstream-validator=vc2_conformance.scripts.vc2_bitstream_validator:main",
            "vc2-test-case-generator=vc2_conformance.scripts.vc2_test_case_generator:main",
            "vc2-test-case-generator-worker=vc2_conformance.scripts.vc2_test_case_generator.worker:main",
            "vc2-picture-explain=vc2_conformance.scripts.vc2_picture_explain:main",
            "vc2-picture-compare=vc2_conformance.scripts.vc2_picture_compare:main",
        ],
    },
)



"""
The :py:mod:`vc2_conformance.decoder.fragment_syntax` module contains pseudocode
functions from (14) Fragment syntax.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.pseudocode.parse_code_functions import using_dc_prediction

from vc2_conformance.decoder.io import (
    tell,
    read_uint_lit,
)

from vc2_conformance.decoder.exceptions import (
    FragmentedPictureRestarted,
    PictureNumberChangedMidFragmentedPicture,
    TooManySlicesInFragmentedPicture,
    FragmentSlicesNotContiguous,
)

from vc2_conformance.decoder.assertions import (
    assert_picture_number_incremented_as_expected,
)

from vc2_conformance.decoder.picture_syntax import transform_parameters

from vc2_conformance.decoder.transform_data_syntax import (
    initialize_wavelet_data,
    slice,
    dc_prediction,
)


__all__ = [
    "fragment_parse",
    "fragment_header",
    "initialize_fragment_state",
    "fragment_data",
]


@ref_pseudocode
def fragment_parse(state):
    """(14.1)"""
    fragment_header(state)
    if state["fragment_slice_count"] == 0:
        transform_parameters(state)
        initialize_fragment_state(state)
    else:
        fragment_data(state)


@ref_pseudocode
def fragment_header(state):
    """(14.2)"""
    fragment_offset = tell(state)  ## Not in spec

    state["picture_number"] = read_uint_lit(state, 4)
    picture_number_offset = tell(state)  ## Not in spec

    state["fragment_data_length"] = read_uint_lit(state, 2)
    state["fragment_slice_count"] = read_uint_lit(state, 2)

    ## Begin not in spec
    if state["fragment_slice_count"] == 0:
        # Errata: not fully defined in spec
        #
        # (14.2) Only the first fragment in a picture may have
        # fragment_slice_count==0 (and no sequence of fragments may be
        # incomplete)
        if state["_fragment_slices_remaining"] != 0:
            raise FragmentedPictureRestarted(
                state["_picture_initial_fragment_offset"],
                fragment_offset,
                state["fragment_slices_received"],
                state["_fragment_slices_remaining"],
            )

        # Errata: wrapping behaviour not constrained in spec
        #
        # (14.2) The picture number should be incremented by one, wrap correctly and
        # fields should have even-numbered picture numbers first.
        assert_picture_number_incremented_as_expected(state, picture_number_offset)

        state["_picture_initial_fragment_offset"] = fragment_offset
    else:
        # (14.2) Appart from when fragment_slice_count==0, the picture number
        # must not change
        if state["_last_picture_number"] != state["picture_number"]:
            raise PictureNumberChangedMidFragmentedPicture(
                state["_last_picture_number_offset"],
                state["_last_picture_number"],
                picture_number_offset,
                state["picture_number"],
            )

        # Errata: not specified in standard...
        #
        # (14.2) A fragmented picture must not contain any extra slices
        if state["fragment_slice_count"] > state["_fragment_slices_remaining"]:
            raise TooManySlicesInFragmentedPicture(
                state["_picture_initial_fragment_offset"],
                fragment_offset,
                state["fragment_slices_received"],
                state["_fragment_slices_remaining"],
                state["fragment_slice_count"],
            )
    ## End not in spec

    if state["fragment_slice_count"] != 0:
        state["fragment_x_offset"] = read_uint_lit(state, 2)
        state["fragment_y_offset"] = read_uint_lit(state, 2)

        # Errata: not specified in standard...
        #
        # (14.2) A fragmented picture must include all slices in raster-scan
        # order
        ## Begin not in spec
        expected_fragment_x_offset = (
            state["fragment_slices_received"] % state["slices_x"]
        )
        expected_fragment_y_offset = (
            state["fragment_slices_received"] // state["slices_x"]
        )
        if (
            state["fragment_x_offset"] != expected_fragment_x_offset
            or state["fragment_y_offset"] != expected_fragment_y_offset
        ):
            raise FragmentSlicesNotContiguous(
                state["_picture_initial_fragment_offset"],
                fragment_offset,
                state["fragment_x_offset"],
                state["fragment_y_offset"],
                expected_fragment_x_offset,
                expected_fragment_y_offset,
            )
        ## End not in spec


@ref_pseudocode
def initialize_fragment_state(state):
    """(14.3)"""
    state["y_transform"] = initialize_wavelet_data(state, "Y")
    state["c1_transform"] = initialize_wavelet_data(state, "C1")
    state["c2_transform"] = initialize_wavelet_data(state, "C2")
    state["fragment_slices_received"] = 0
    ## Begin not in spec
    state["_fragment_slices_remaining"] = state["slices_x"] * state["slices_y"]
    ## End not in spec
    state["fragmented_picture_done"] = False


@ref_pseudocode
def fragment_data(state):
    """(14.4)"""
    for s in range(state["fragment_slice_count"]):
        slice_x = (
            state["fragment_y_offset"] * state["slices_x"]
            + state["fragment_x_offset"]
            + s
        ) % state["slices_x"]
        slice_y = (
            state["fragment_y_offset"] * state["slices_x"]
            + state["fragment_x_offset"]
            + s
        ) // state["slices_x"]
        slice(state, slice_x, slice_y)
        state["fragment_slices_received"] += 1
        state["_fragment_slices_remaining"] -= 1  ## Not in spec
        if state["fragment_slices_received"] == state["slices_x"] * state["slices_y"]:
            state["fragmented_picture_done"] = True
            if using_dc_prediction(state):
                if state["dwt_depth_ho"] == 0:
                    dc_prediction(state["y_transform"][0]["LL"])
                    dc_prediction(state["c1_transform"][0]["LL"])
                    dc_prediction(state["c2_transform"][0]["LL"])
                else:
                    dc_prediction(state["y_transform"][0]["L"])
                    dc_prediction(state["c1_transform"][0]["L"])
                    dc_prediction(state["c2_transform"][0]["L"])



"""
:py:mod:`vc2_conformance.bitstream.exceptions`
==============================================

Custom exception types related to bitstream parsing/serialisation.
"""


class OutOfRangeError(ValueError):
    """
    An exception thrown whenever an out-of-range value is passed to a bitstream
    writing function.
    """


class UnusedTargetError(ValueError):
    """
    Thrown by functions in :py:mod:`vc2_conformance.bitstream` when a
    value in a context dictionary was left used.
    """


class ReusedTargetError(ValueError):
    """
    Thrown by functions in :py:mod:`vc2_conformance.bitstream` when a
    target in a context dictionary is used more than once.
    """


class ListTargetExhaustedError(IndexError):
    """
    Thrown by functions in :py:mod:`vc2_conformance.bitstream` when a value
    beyond the end of a list target in a context dictionary is requested.
    """


class ListTargetContainsNonListError(ValueError):
    """
    Thrown by functions in :py:mod:`vc2_conformance.bitstream` when a
    target in a context dictionary has been declared to be a list but does not
    contain a value of the :py:class:`list` type.
    """


class UnclosedBoundedBlockError(ValueError):
    """
    Thrown by functions in :py:mod:`vc2_conformance.bitstream` when some
    :py:meth:`SerDes.bounded_block_begin` does not have corresponding
    :py:meth:`SerDes.bounded_block_end`.
    """


class UnclosedNestedContextError(ValueError):
    """
    Thrown by functions in :py:mod:`vc2_conformance.bitstream` when some
    :py:meth:`SerDes.subcontext_enter` does not have corresponding
    :py:meth:`SerDes.subcontext_leave`.
    """



"""
The :py:mod:`vc2_conformance.encoder.sequence_header` module contains routines
for encoding a set of video format and codec parameters into sequence headers.

The :py:func:`make_sequence_header_data_unit` function is used to generate
sequence headers by the encoder:

.. autofunction:: make_sequence_header_data_unit

In practice there are often many potential sequence header encodings for a
given set of video parameters. For example, when a video format closely matches
a predefined base video format, the various ``custom_*_flag`` overrides may
largely be omitted. This is optional, however, and an encoder is free to use
these overrides explicitly even when they're not required.

The :py:func:`make_sequence_header_data_unit` function always attempts to use
the most compact encoding it can. Some test cases, however may wish to use less
compact encodings and so to support this the :py:func:`iter_sequence_headers`
function is provided:

.. autofunction:: iter_sequence_headers

"""

from functools import partial

from vc2_data_tables import (
    Levels,
    ParseCodes,
    BaseVideoFormats,
    BASE_VIDEO_FORMAT_PARAMETERS,
    PRESET_FRAME_RATES,
    PRESET_PIXEL_ASPECT_RATIOS,
    PRESET_SIGNAL_RANGES,
    PRESET_COLOR_SPECS,
)

from vc2_conformance.constraint_table import (
    filter_constraint_table,
    allowed_values_for,
)

from vc2_conformance.level_constraints import (
    LEVEL_CONSTRAINTS,
    LEVEL_CONSTRAINT_ANY_VALUES,
)

from vc2_conformance.codec_features import codec_features_to_trivial_level_constraints

from vc2_conformance.pseudocode.video_parameters import set_source_defaults

from vc2_conformance.bitstream import (
    DataUnit,
    ParseInfo,
    ParseParameters,
    SequenceHeader,
    SourceParameters,
    FrameSize,
    ColorDiffSamplingFormat,
    ScanFormat,
    FrameRate,
    PixelAspectRatio,
    CleanArea,
    SignalRange,
    ColorSpec,
    ColorPrimaries,
    ColorMatrix,
    TransferFunction,
)

from vc2_conformance.encoder.exceptions import IncompatibleLevelAndVideoFormatError


__all__ = [
    "rank_base_video_format_similarity",
    "rank_allowed_base_video_format_similarity",
    "iter_source_parameter_options",
    "make_parse_parameters",
    "iter_sequence_headers",
    "make_sequence_header",
    "make_sequence_header_data_unit",
]


def zip_longest_repeating_final_value(*iterables):
    """
    Like :py:func:`zip`, but if some iterators finish before others, their last
    value is repeated to fill in the missing entries.

    For example::

        >>> list(zip_longest_repeat_last(
        ...     [1, 2, 3],
        ...     ["a", "b", "c", "d", "e"],
        ... ))
        [(1, "a"), (2, "b"), (3, "c"), (3, "d"), (3, "e")]

    If an iterator doesn't produce any values, its missing values will be
    filled with None.
    """
    iterators = list(map(iter, iterables))

    last_zipped = (None,) * len(iterators)

    some_iterators_running = True
    while some_iterators_running:
        zipped = []
        some_iterators_running = False
        for i, iterator in enumerate(iterators):
            try:
                zipped.append(next(iterator))
                some_iterators_running = True
            except StopIteration:
                zipped.append(last_zipped[i])

        if some_iterators_running:
            last_zipped = tuple(zipped)
            yield last_zipped


def iter_custom_options_dicts(
    base_video_parameters,
    video_parameters,
    level_constraints_dict,
    dict_type,
    flag_key,
    parameters,
    presets=None,
    preset_index_constraint_key=None,
):
    """
    A generator which enumerates valid sets of VC-2 custom option settings
    which achieve a set of video parameters given a set of base video
    parameters and level constraints.

    Options dictionaries will be generated in ascending order of explicitness.

    Example usage::

        >>> from functools import partial
        >>> from collections import defaultdict
        >>> from vc2_data_tables import (
        ...     BASE_VIDEO_FORMAT_PARAMETERS,
        ...     PRESET_PIXEL_ASPECT_RATIOS,
        ... )
        >>> from vc2_conformance.constraint_table import AnyValue
        >>> from vc2_conformance.bitstream import (
        ...     PixelAspectRatio,
        ... )
        >>> from vc2_conformance.pseudocode.video_parameters import (
        ...     PixelAspectRatio,
        ...     set_source_defaults,
        ... )

        >>> iter_pixel_aspect_ratio_options = partial(
        ...     iter_custom_options_dicts,
        ...     dict_type=PixelAspectRatio,
        ...     flag_key="custom_pixel_aspect_ratio_flag",
        ...     parameters=[
        ...         "pixel_aspect_ratio_numer",
        ...         "pixel_aspect_ratio_denom",
        ...     ],
        ...     PRESET_PIXEL_ASPECT_RATIOS,
        ... )

        >>> # In this case, the format matches the default so can use any type
        >>> # of override
        >>> base_video_parameters = set_source_defaults(
        ...     BASE_VIDEO_FORMAT_PARAMETERS[
        ...         BaseVideoFormats.sdi576i_50
        ...     ]
        ... )
        >>> video_parameters = VideoParameters(
        ...     pixel_aspect_ratio_numer=12,
        ...     pixel_aspect_ratio_denom=11,
        ...     # ...
        ... )
        >>> level_constraints_dict = defaultdict(AnyValue)  # ...for sake of example
        >>> for f in iter_pixel_aspect_ratio_options(
        ...     base_video_parameters,
        ...     video_parameters,
        ...     level_constraints_dict,
        ... ):
        ...     print(f)
        PixelAspectRatio:
          custom_pixel_aspect_ratio_flag: False
        PixelAspectRatio:
          custom_pixel_aspect_ratio_flag: True
          index: ratio_12_11 (3)
        PixelAspectRatio:
          custom_pixel_aspect_ratio_flag: True
          index: 0
          pixel_aspect_ratio_numer: 12
          pixel_aspect_ratio_denom: 11

        >>> # In this case, the format doesn't match the base video format so
        >>> # we have to override
        >>> video_parameters = VideoParameters(
        ...     pixel_aspect_ratio_numer=4,
        ...     pixel_aspect_ratio_denom=3,
        ...     # ...
        ... )
        >>> for f in iter_pixel_aspect_ratio_options(
        ...     base_video_parameters,
        ...     video_parameters,
        ...     level_constraints_dict,
        ... ):
        ...     print(f)
        PixelAspectRatio:
          custom_pixel_aspect_ratio_flag: True
          index: reduced_horizontal_resolution (6)
        PixelAspectRatio:
          custom_pixel_aspect_ratio_flag: True
          index: 0
          pixel_aspect_ratio_numer: 4
          pixel_aspect_ratio_denom: 3

    Parameters
    ==========
    base_video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        representing the current base video parameters. See
        :py:func:`~vc2_conformance.pseudocode.video_parameters.set_source_defaults`.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        representing the desired video parameters to be represented. Only the
        entries mentioned in ``dict_type`` will be checked.
    level_constraints_dict : {str: :py:class:`~vc2_conformance.constraint_table.ValueSet`, ...}
        A single dictionary of level constraints (i.e. a single column from
        :py:class:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS`).
    dict_type : :py:class:`~vc2_conformance.fixeddicts.fixeddict` type
        The type of :py:class:`~vc2_conformance.fixeddicts.fixeddict` which
        holds the required custom option.
    flag_key : str
        Name of the ``custom_*_flag`` entry in ``dict_type``.
    parameters : [str or (str, str), ...]
        Names of keys in ``dict_type`` to try and match in the
        video_parameters. If the dict_type and VideoParameters field names
        differ, a pair (video_parameters_name, dict_type_name) should be given
        instead.
    presets : {index: namedtuple(value, ...), ...} or None
        If present, gives the set of default values for this ``dict_type``
        which would be indicated by a field called 'index'. If this argument is
        None, no 'index' field will be added. The names in ``parameters`` must
        match the order in the presets.
    preset_index_constraint_key : str
        Required if ``presets`` is not None. The name of the
        :py:class:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS` key
        which corresponds with the index field. Ignored otherwise.
    """
    # Normalise to (vp_key, dt_key) pairs
    parameters = [(key, key) if isinstance(key, str) else key for key in parameters]

    # Is the base video format already sufficient (and are we allowed to leave
    # the flag cleared)?
    if (
        all(
            base_video_parameters[vp_key] == video_parameters[vp_key]
            for vp_key, _ in parameters
        )
        and False in level_constraints_dict[flag_key]
    ):
        yield dict_type({flag_key: False})

    # Is there a suitable preset (and are we allowed to set it)?
    if presets is not None:
        for index, values in presets.items():
            if values == tuple(video_parameters[vp_key] for vp_key, _ in parameters):
                if (
                    True in level_constraints_dict[flag_key]
                    and index in level_constraints_dict[preset_index_constraint_key]
                ):
                    yield dict_type({flag_key: True, "index": index})

    # Explicitly set the desired value (if we're allowed to do so)
    if (
        True in level_constraints_dict[flag_key]
        and (
            presets is None or 0 in level_constraints_dict[preset_index_constraint_key]
        )
        and all(
            # NB: Constraints table entry names are shared with the
            # VideoParameters keys (and not the pseudocode/bitstream fixeddicts
            # whose names are not globally unique).
            video_parameters[vp_key] in level_constraints_dict[vp_key]
            for vp_key, dt_key in parameters
        )
    ):
        out = dict_type({flag_key: True})
        if presets is not None:
            out["index"] = 0
        for vp_key, dt_key in parameters:
            out[dt_key] = video_parameters[vp_key]
        yield out


iter_frame_size_options = partial(
    iter_custom_options_dicts,
    dict_type=FrameSize,
    flag_key="custom_dimensions_flag",
    parameters=["frame_width", "frame_height"],
)

iter_color_diff_sampling_format_options = partial(
    iter_custom_options_dicts,
    dict_type=ColorDiffSamplingFormat,
    flag_key="custom_color_diff_format_flag",
    parameters=["color_diff_format_index"],
)

iter_scan_format_options = partial(
    iter_custom_options_dicts,
    dict_type=ScanFormat,
    flag_key="custom_scan_format_flag",
    parameters=["source_sampling"],
)

iter_frame_rate_options = partial(
    iter_custom_options_dicts,
    dict_type=FrameRate,
    flag_key="custom_frame_rate_flag",
    parameters=["frame_rate_numer", "frame_rate_denom"],
    presets=PRESET_FRAME_RATES,
    preset_index_constraint_key="frame_rate_index",
)

iter_pixel_aspect_ratio_options = partial(
    iter_custom_options_dicts,
    dict_type=PixelAspectRatio,
    flag_key="custom_pixel_aspect_ratio_flag",
    parameters=["pixel_aspect_ratio_numer", "pixel_aspect_ratio_denom"],
    presets=PRESET_PIXEL_ASPECT_RATIOS,
    preset_index_constraint_key="pixel_aspect_ratio_index",
)

iter_clean_area_options = partial(
    iter_custom_options_dicts,
    dict_type=CleanArea,
    flag_key="custom_clean_area_flag",
    parameters=["clean_width", "clean_height", "top_offset", "left_offset"],
)

iter_signal_range_options = partial(
    iter_custom_options_dicts,
    dict_type=SignalRange,
    flag_key="custom_signal_range_flag",
    parameters=[
        "luma_offset",
        "luma_excursion",
        "color_diff_offset",
        "color_diff_excursion",
    ],
    presets=PRESET_SIGNAL_RANGES,
    preset_index_constraint_key="custom_signal_range_index",
)

iter_color_primaries_options = partial(
    iter_custom_options_dicts,
    dict_type=ColorPrimaries,
    flag_key="custom_color_primaries_flag",
    parameters=[("color_primaries_index", "index")],
)

iter_color_matrix_options = partial(
    iter_custom_options_dicts,
    dict_type=ColorMatrix,
    flag_key="custom_color_matrix_flag",
    parameters=[("color_matrix_index", "index")],
)

iter_transfer_function_options = partial(
    iter_custom_options_dicts,
    dict_type=TransferFunction,
    flag_key="custom_transfer_function_flag",
    parameters=[("transfer_function_index", "index")],
)


def iter_color_spec_options(
    base_video_parameters, video_parameters, level_constraints_dict
):
    """
    Generates a series of :py:class:`vc2_conformance.bitstream.ColorSpec`
    dictionaries which may be used to achieve the specified video parameters
    while obeying the provided level constraints.

    Where defaults are permissible, examples will be produced with each flag
    set to both True and False, though not every combination of flags will be
    given.

    Parameters
    ==========
    base_video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        representing the current base video parameters. See
        :py:func:`~vc2_conformance.pseudocode.video_parameters.set_source_defaults`.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        representing the desired video parameters to be represented. Only the
        color mode related entries mentioned in will be checked.
    level_constraints_dict : {str: :py:class:`~vc2_conformance.constraint_table.ValueSet`, ...}
        A single dictionary of level constraints (i.e. a single column from
        :py:class:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS`).
    """
    # Is this already covered by the base format (and are we allowed to use it)?
    if (
        all(
            base_video_parameters[key] == video_parameters[key]
            for key in [
                "color_primaries_index",
                "color_matrix_index",
                "transfer_function_index",
            ]
        )
        and False in level_constraints_dict["custom_color_spec_flag"]
    ):
        yield ColorSpec(
            custom_color_spec_flag=False,
        )

    # Can a preset (other than 0) satisfy the requirement (and are we allowed
    # to use it)?
    for index, (
        primaries,
        matrix,
        tf,
    ) in PRESET_COLOR_SPECS.items():
        if (
            index != 0
            and video_parameters["color_primaries_index"] == primaries
            and video_parameters["color_matrix_index"] == matrix
            and video_parameters["transfer_function_index"] == tf
            and True in level_constraints_dict["custom_color_spec_flag"]
            and index in level_constraints_dict["color_spec_index"]
        ):
            yield ColorSpec(
                custom_color_spec_flag=True,
                index=index,
            )

    # Work through all possible ways to express this using fully custom formats
    # (if we're allowed to do so)
    custom_base_vp = base_video_parameters.copy()
    custom_presets = PRESET_COLOR_SPECS[0]
    custom_base_vp["color_primaries_index"] = custom_presets.color_primaries_index
    custom_base_vp["color_matrix_index"] = custom_presets.color_matrix_index
    custom_base_vp["transfer_function_index"] = custom_presets.transfer_function_index

    if (
        True in level_constraints_dict["custom_color_spec_flag"]
        and 0 in level_constraints_dict["color_spec_index"]
    ):
        for (
            color_primaries,
            color_matrix,
            transfer_function,
        ) in zip_longest_repeating_final_value(
            iter_color_primaries_options(
                custom_base_vp,
                video_parameters,
                level_constraints_dict,
            ),
            iter_color_matrix_options(
                custom_base_vp,
                video_parameters,
                level_constraints_dict,
            ),
            iter_transfer_function_options(
                custom_base_vp,
                video_parameters,
                level_constraints_dict,
            ),
        ):
            # Give up if we're unable to produce a needed custom color option
            if (
                color_primaries is None
                or color_matrix is None
                or transfer_function is None
            ):
                break

            yield ColorSpec(
                custom_color_spec_flag=True,
                index=0,
                color_primaries=color_primaries,
                color_matrix=color_matrix,
                transfer_function=transfer_function,
            )


def iter_source_parameter_options(
    base_video_parameters, video_parameters, level_constraints_dict
):
    """
    Generates a series of
    :py:class:`vc2_conformance.bitstream.SourceParameters` dictionaries which
    may be used to set the specified ``video_parameters``, starting from the
    supplied ``base_video_parameters``.

    Parameters
    ==========
    base_video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        representing the current base video parameters. See
        :py:func:`~vc2_conformance.pseudocode.video_parameters.set_source_defaults`.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        representing the desired video parameters to be represented.
    level_constraints_dict : {str: :py:class:`~vc2_conformance.constraint_table.ValueSet`, ...}
        A single dictionary of level constraints (i.e. a single column from
        :py:class:`vc2_conformance.level_constraints.LEVEL_CONSTRAINTS`). All
        generated parameter sets will meet the restrictions imposed (which may
        include preventing any cases being generated).

    Yields
    ======
    source_parameters : :py:class:`vc2_conformance.bitstream.SourceParameters`
        A series of :py:class:`vc2_conformance.bitstream.SourceParameters`
        dictionaries, starting with the simplest (fewest custom overrides
        possible), building up to the most explicit (all options have custom
        overrides).
    """
    # Special case: (11.3) The top field first parameter cannot be overriden
    # from the base video format. If these are mismatched, we cannot produce
    # any useful results.
    if base_video_parameters["top_field_first"] != video_parameters["top_field_first"]:
        return

    for (
        frame_size,
        color_diff_sampling_format,
        scan_format,
        frame_rate,
        pixel_aspect_ratio,
        clean_area,
        signal_range,
        color_spec,
    ) in zip_longest_repeating_final_value(
        *(
            fn(base_video_parameters, video_parameters, level_constraints_dict)
            for fn in [
                iter_frame_size_options,
                iter_color_diff_sampling_format_options,
                iter_scan_format_options,
                iter_frame_rate_options,
                iter_pixel_aspect_ratio_options,
                iter_clean_area_options,
                iter_signal_range_options,
                iter_color_spec_options,
            ]
        )
    ):
        source_parameters = SourceParameters(
            frame_size=frame_size,
            color_diff_sampling_format=color_diff_sampling_format,
            scan_format=scan_format,
            frame_rate=frame_rate,
            pixel_aspect_ratio=pixel_aspect_ratio,
            clean_area=clean_area,
            signal_range=signal_range,
            color_spec=color_spec,
        )

        # Give up immediately if any of the parts cannot be satisifed
        if any(value is None for value in source_parameters.values()):
            break

        yield source_parameters


def count_video_parameter_differences(a, b):
    """
    Given two sets of :py:class:`VideoParameters`, return a count of the values
    which differ.
    """
    # NB: Should always be the same in practice...
    all_keys = set(a) | set(b)

    differences = 0
    for key in all_keys:
        if key not in a:
            differences += 1
        elif key not in b:
            differences += 1
        elif a[key] != b[key]:
            differences += 1

    return differences


def rank_base_video_format_similarity(
    video_parameters,
    base_video_formats=list(BaseVideoFormats),
):
    """
    Given a set of
    :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`, return an
    ordered list of :py:class:`~vc2_data_tables.BaseVideoFormats` with the most
    similar format first and least similar last.

    .. note::
        The returned :py:class:`~vc2_data_tables.BaseVideoFormats` will always
        have the same ``top_field_first`` setting.

    Parameters
    ==========
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The video parameters against which to rank base video formats'
        similarity.
    base_video_formats : [:py:class:`~vc2_data_tables.BaseVideoFormats`, ...]
        Optional. The base video format indices to consider. Defaults to all
        base video formats.
    """
    return sorted(
        (
            index
            for index in base_video_formats
            if (
                BASE_VIDEO_FORMAT_PARAMETERS[index].top_field_first
                == video_parameters["top_field_first"]
            )
        ),
        key=lambda index: count_video_parameter_differences(
            set_source_defaults(index),
            video_parameters,
        ),
    )


def rank_allowed_base_video_format_similarity(codec_features):
    """
    Produce a ranked list of base video format IDs for the provided codec
    features. Video formats with the wrong 'top field first' value and which
    aren't allowed by the current level are omitted. The most similar base
    format is returned first.

    Raises
    :py:exc:`~vc2_conformance.encoder.exceptions.IncompatibleLevelAndVideoFormatError`
    if no suitable base video format is available.
    """

    constrained_values = codec_features_to_trivial_level_constraints(codec_features)

    return rank_base_video_format_similarity(
        codec_features["video_parameters"],
        list(
            allowed_values_for(
                LEVEL_CONSTRAINTS,
                "base_video_format",
                constrained_values,
                LEVEL_CONSTRAINT_ANY_VALUES["base_video_format"],
            ).iter_values()
        ),
    )


def make_parse_parameters(codec_features):
    """
    Create a :py:class:`~vc2_conformance.bitstream.ParseParameters` object
    using the profile and level specified in the
    :py:class:`~vc2_conformance.codec_features.CodecFeatures` dictionary
    provided.
    """
    return ParseParameters(
        profile=codec_features["profile"],
        level=codec_features["level"],
    )


def iter_sequence_headers(codec_features):
    """
    Generate a series of :py:class:`~vc2_conformance.bitstream.SequenceHeader`
    objects which encode the video format specified in
    :py:class:`~vc2_conformance.codec_features.CodecFeatures` dictionary
    provided.

    This generator will start with an efficient encoding of the required
    features, built on the most closely matched base video format. This will be
    followed by successively less efficient encodings (i.e. using more custom
    fields) but the same (best-matched) base video format. After this,
    encodings based on other base video formats will be produced (again
    starting with the most efficient encoding for each format first).

    This generator may output no items if the VC-2 level specified does not
    permit the format given.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`

    Yields
    ======
    sequence_header : :py:class:`~vc2_conformance.bitstream.SequenceHeader`
    """
    picture_coding_mode = codec_features["picture_coding_mode"]
    video_parameters = codec_features["video_parameters"]

    constrained_values = codec_features_to_trivial_level_constraints(codec_features)

    # Level constraints may force us to use a particular base video format for
    # this encoding so we try all possible encodings (starting with the most
    # compact) and stop when we find one compliant with the level restrictions.
    base_video_formats = rank_allowed_base_video_format_similarity(codec_features)
    for base_video_format in base_video_formats:
        base_video_parameters = set_source_defaults(base_video_format)

        filtered_constraint_table = filter_constraint_table(
            LEVEL_CONSTRAINTS,
            dict(constrained_values, base_video_format=base_video_format),
        )
        for level_constraints_dict in filtered_constraint_table:
            for source_parameters in iter_source_parameter_options(
                base_video_parameters, video_parameters, level_constraints_dict
            ):
                yield SequenceHeader(
                    parse_parameters=make_parse_parameters(codec_features),
                    base_video_format=base_video_format,
                    video_parameters=source_parameters,
                    picture_coding_mode=picture_coding_mode,
                )


def make_sequence_header(codec_features):
    """
    Create a :py:class:`~vc2_conformance.bitstream.SequenceHeader` object which
    efficiently encodes the video format specified in
    :py:class:`~vc2_conformance.codec_features.CodecFeatures` dictionary
    provided.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`

    Returns
    =======
    sequence_header : :py:class:`~vc2_conformance.bitstream.SequenceHeader`

    Raises
    =======
    :py:exc:`IncompatibleLevelAndVideoFormatError`
    """
    try:
        return next(iter_sequence_headers(codec_features))
    except StopIteration:
        # No suitable encoding was possible. This (should!) only occur due to level
        # constraints so we'll blame these here.
        #
        # This assertion (should be) unreachable and is here so that we don't
        # misleadingly blame levels in the event of this code being broken!
        assert codec_features["level"] != Levels.unconstrained
        raise IncompatibleLevelAndVideoFormatError(codec_features)


def make_sequence_header_data_unit(codec_features):
    """
    Create a :py:class:`~vc2_conformance.bitstream.DataUnit` object containing
    a sequence header which sensibly encodes the features specified in
    :py:class:`~vc2_conformance.codec_features.CodecFeatures`
    dictionary provided.

    Parameters
    ==========
    codec_features : :py:class:`~vc2_conformance.codec_features.CodecFeatures`

    Returns
    =======
    data_unit : :py:class:`~vc2_conformance.bitstream.DataUnit`

    Raises
    =======
    :py:exc:`IncompatibleLevelAndVideoFormatError`
    """
    return DataUnit(
        parse_info=ParseInfo(parse_code=ParseCodes.sequence_header),
        sequence_header=make_sequence_header(codec_features),
    )



"""
Video parameter computation functions (11)

The functions in this module make up the purely functional (i.e. non
bitstream-reading) logic for computing video parameters. This predominantly
consists of preset-loading and component dimension calculation routines.

For functions below which take a 'state' argument, this should be a
dictionary-like object (e.g.
:py:class:`~vc2_conformance.pseudocode.state.State`) with the following
entries:

* ``"frame_width"``
* ``"frame_height"``
* ``"luma_width"``
* ``"luma_height"``
* ``"color_diff_width"``
* ``"color_diff_height"``
* ``"luma_depth"``
* ``"color_diff_depth"``
* ``"picture_coding_mode"``
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.fixeddict import fixeddict, Entry

from vc2_data_tables import (
    PictureCodingModes,
    ColorDifferenceSamplingFormats,
    SourceSamplingModes,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
    BASE_VIDEO_FORMAT_PARAMETERS,
    PRESET_FRAME_RATES,
    PRESET_PIXEL_ASPECT_RATIOS,
    PRESET_SIGNAL_RANGES,
    PRESET_COLOR_SPECS,
)

from vc2_conformance.pseudocode.vc2_math import intlog2

__all__ = [
    "VideoParameters",
    "set_source_defaults",
    "set_coding_parameters",
    "picture_dimensions",
    "video_depth",
    "preset_frame_rate",
    "preset_pixel_aspect_ratio",
    "preset_signal_range",
    "preset_color_primaries",
    "preset_color_matrix",
    "preset_transfer_function",
    "preset_color_spec",
]


VideoParameters = fixeddict(
    "VideoParameters",
    # (11.4.3) frame_size
    Entry("frame_width", help_type="int", help="Set by (11.4.3) frame_size"),
    Entry("frame_height", help_type="int", help="Set by (11.4.3) frame_size"),
    # (11.4.4) color_diff_sampling_format
    Entry(
        "color_diff_format_index",
        enum=ColorDifferenceSamplingFormats,
        help_type=":py:class:`~vc2_data_tables.ColorDifferenceSamplingFormats`",
        help="Set by (11.4.4) color_diff_sampling_format",
    ),
    # (11.4.5) scan_format
    Entry(
        "source_sampling",
        enum=SourceSamplingModes,
        help_type=":py:class:`~vc2_data_tables.SourceSamplingModes`",
        help="Set by (11.4.5) scan_format",
    ),
    # (11.4.5)
    Entry(
        "top_field_first",
        help_type="bool",
        help="Set by (11.4.5) set_source_defaults",
    ),
    # (11.4.6) frame_rate
    Entry("frame_rate_numer", help_type="int", help="Set by (11.4.6) frame_rate"),
    Entry("frame_rate_denom", help_type="int", help="Set by (11.4.6) frame_rate"),
    # (11.4.7) aspect_ratio
    Entry(
        "pixel_aspect_ratio_numer", help_type="int", help="Set by (11.4.7) aspect_ratio"
    ),
    Entry(
        "pixel_aspect_ratio_denom", help_type="int", help="Set by (11.4.7) aspect_ratio"
    ),
    # (11.4.8) clean_area
    Entry("clean_width", help_type="int", help="Set by (11.4.8) clean_area"),
    Entry("clean_height", help_type="int", help="Set by (11.4.8) clean_area"),
    Entry("left_offset", help_type="int", help="Set by (11.4.8) clean_area"),
    Entry("top_offset", help_type="int", help="Set by (11.4.8) clean_area"),
    # (11.4.9) signal_range
    Entry("luma_offset", help_type="int", help="Set by (11.4.9) signal_range"),
    Entry("luma_excursion", help_type="int", help="Set by (11.4.9) signal_range"),
    Entry("color_diff_offset", help_type="int", help="Set by (11.4.9) signal_range"),
    Entry("color_diff_excursion", help_type="int", help="Set by (11.4.9) signal_range"),
    # (11.4.10.2) color_primaries
    Entry(
        "color_primaries_index",
        enum=PresetColorPrimaries,
        help_type=":py:class:`~vc2_data_tables.PresetColorPrimaries`",
        help="Set by (11.4.10.2) color_primaries",
    ),
    # (11.4.10.3) color_matrix
    Entry(
        "color_matrix_index",
        enum=PresetColorMatrices,
        help_type=":py:class:`~vc2_data_tables.PresetColorMatrices`",
        help="Set by (11.4.10.3) color_matrix",
    ),
    # (11.4.10.4) transfer_function
    Entry(
        "transfer_function_index",
        enum=PresetTransferFunctions,
        help_type=":py:class:`~vc2_data_tables.PresetTransferFunctions`",
        help="Set by (11.4.10.4) transfer_function",
    ),
    help="""
        (11.4) Video parameters struct.
    """,
)


@ref_pseudocode(deviation="inferred_implementation")
def set_source_defaults(base_video_format):
    """
    (11.4.2) Create a :py:class:`VideoParameters` object with the parameters
    specified in a base video format.
    """
    base = BASE_VIDEO_FORMAT_PARAMETERS[base_video_format]
    preset_frame_rates = PRESET_FRAME_RATES[base.frame_rate_index]
    preset_pixel_aspect_ratios = PRESET_PIXEL_ASPECT_RATIOS[
        base.pixel_aspect_ratio_index
    ]
    preset_signal_ranges = PRESET_SIGNAL_RANGES[base.signal_range_index]
    preset_color_spec = PRESET_COLOR_SPECS[base.color_spec_index]
    return VideoParameters(
        frame_width=base.frame_width,
        frame_height=base.frame_height,
        color_diff_format_index=base.color_diff_format_index,
        source_sampling=base.source_sampling,
        top_field_first=base.top_field_first,
        frame_rate_numer=preset_frame_rates.numerator,
        frame_rate_denom=preset_frame_rates.denominator,
        pixel_aspect_ratio_numer=preset_pixel_aspect_ratios.numerator,
        pixel_aspect_ratio_denom=preset_pixel_aspect_ratios.denominator,
        clean_width=base.clean_width,
        clean_height=base.clean_height,
        left_offset=base.left_offset,
        top_offset=base.top_offset,
        luma_offset=preset_signal_ranges.luma_offset,
        luma_excursion=preset_signal_ranges.luma_excursion,
        color_diff_offset=preset_signal_ranges.color_diff_offset,
        color_diff_excursion=preset_signal_ranges.color_diff_excursion,
        color_primaries_index=preset_color_spec.color_primaries_index,
        color_matrix_index=preset_color_spec.color_matrix_index,
        transfer_function_index=preset_color_spec.transfer_function_index,
    )


@ref_pseudocode
def set_coding_parameters(state, video_parameters):
    """(11.6.1) Set picture coding mode parameter."""
    picture_dimensions(state, video_parameters)
    video_depth(state, video_parameters)


@ref_pseudocode
def picture_dimensions(state, video_parameters):
    """(11.6.2) Compute the picture component dimensions in global state."""
    state["luma_width"] = video_parameters["frame_width"]
    state["luma_height"] = video_parameters["frame_height"]
    state["color_diff_width"] = state["luma_width"]
    state["color_diff_height"] = state["luma_height"]

    if (
        video_parameters["color_diff_format_index"]
        == ColorDifferenceSamplingFormats.color_4_2_2
    ):
        state["color_diff_width"] //= 2
    if (
        video_parameters["color_diff_format_index"]
        == ColorDifferenceSamplingFormats.color_4_2_0
    ):
        state["color_diff_width"] //= 2
        state["color_diff_height"] //= 2

    if state["picture_coding_mode"] == PictureCodingModes.pictures_are_fields:
        state["luma_height"] //= 2
        state["color_diff_height"] //= 2


@ref_pseudocode
def video_depth(state, video_parameters):
    """(11.6.3) Compute the bits-per-sample for the decoded video."""
    state["luma_depth"] = intlog2(video_parameters["luma_excursion"] + 1)
    state["color_diff_depth"] = intlog2(video_parameters["color_diff_excursion"] + 1)


@ref_pseudocode(deviation="inferred_implementation")
def preset_frame_rate(video_parameters, index):
    """(11.4.6) Set frame rate from preset."""
    preset = PRESET_FRAME_RATES[index]
    video_parameters["frame_rate_numer"] = preset.numerator
    video_parameters["frame_rate_denom"] = preset.denominator


@ref_pseudocode(deviation="inferred_implementation")
def preset_pixel_aspect_ratio(video_parameters, index):
    """(11.4.7) Set pixel aspect ratio from preset."""
    preset = PRESET_PIXEL_ASPECT_RATIOS[index]
    video_parameters["pixel_aspect_ratio_numer"] = preset.numerator
    video_parameters["pixel_aspect_ratio_denom"] = preset.denominator


@ref_pseudocode(deviation="inferred_implementation")
def preset_signal_range(video_parameters, index):
    """(11.4.7) Set signal range from preset."""
    preset = PRESET_SIGNAL_RANGES[index]
    video_parameters["luma_offset"] = preset.luma_offset
    video_parameters["luma_excursion"] = preset.luma_excursion
    video_parameters["color_diff_offset"] = preset.color_diff_offset
    video_parameters["color_diff_excursion"] = preset.color_diff_excursion


@ref_pseudocode(deviation="inferred_implementation")
def preset_color_primaries(video_parameters, index):
    """(11.4.10.2) Set the color primaries parameter from a preset."""
    video_parameters["color_primaries_index"] = index


@ref_pseudocode(deviation="inferred_implementation")
def preset_color_matrix(video_parameters, index):
    """(11.4.10.3) Set the color matrix parameter from a preset."""
    video_parameters["color_matrix_index"] = index


@ref_pseudocode(deviation="inferred_implementation")
def preset_transfer_function(video_parameters, index):
    """(11.4.10.4) Set the transfer function parameter from a preset."""
    video_parameters["transfer_function_index"] = index


@ref_pseudocode(deviation="inferred_implementation")
def preset_color_spec(video_parameters, index):
    """(11.4.10.1) Load a preset color specification."""
    preset = PRESET_COLOR_SPECS[index]
    preset_color_primaries(video_parameters, preset.color_primaries_index)
    preset_color_matrix(video_parameters, preset.color_matrix_index)
    preset_transfer_function(video_parameters, preset.transfer_function_index)



"""
The :py:mod:`vc2_conformance.color_conversion` module implements color system
related functions relating to the color formats supported by VC-2.

The primary use for this module is to provide routines for converting between
colors specified by VC-2's various supported color systems (see Annex (E.1) of
the VC-2 specification). This functionality is used during the generation of
certain encoder and decoder test cases (:py:mod:`vc2_conformance.test_cases`).

High-level API
--------------

This module implements simple color format conversion routines for converting
between arbitrary VC-2 color formats via floating point CIE XYZ color. The
process is implemented by the following high-level functions:

.. autofunction:: to_xyz

.. autofunction:: from_xyz

.. warning::

    Color format conversion is an extremely complex problem. The approach used
    by this module is simplistic in both its approach and implementation. While
    it will always produce plausible colors, it may not produce the best
    possible result. To give a few examples of limitations of this module:

    * Potential numerical stability issues are ignored (e.g. YCgCo conversions
      may be lossy)
    * No white point correction is applied
    * Out-of-gamut colors are crudely clipped
    * Poor quality antialiasing filters for chroma subsampling/interpolation

    Finally, this module should be considered a 'best effort' at a correct
    implementation and the resulting color conversion should largely be
    treated as informative.

.. warning::

    Support :py:func:`to_xyz` is limited to only formats using the
    :py:data:`~vc2_data_tables.PresetTransferFunctions.tv_gamma` transfer
    function. All formats are supported, however, by :py:func:`from_xyz`.


Low-level API
-------------

The conversion processes used by :py:func:`to_xyz` and :py:func:`from_xyz` is
built on a series of lower-level transformations as described by the figure
below. These lower-level primitives may be used directly to perform more
specialised conversions.

      .. image:: /_static/color_conversion_process.svg

These steps build on the following conversion functions and matrices. These are
implemented based on the specifications cited by the VC-2 specification.

.. autofunction:: float_to_int_clipped

.. autofunction:: float_to_int

.. autofunction:: int_to_float

.. autofunction:: from_444

.. autofunction:: to_444

.. autodata:: COLOR_MATRICES
    :annotation: = {<color matrix index>: <3x3 matrix>, ...}

.. autodata:: INVERSE_COLOR_MATRICES
    :annotation: = {<color matrix index>: <3x3 matrix>, ...}

.. autodata:: TRANSFER_FUNCTIONS
    :annotation: = {<transfer function index>: <function>, ...}

.. autodata:: INVERSE_TRANSFER_FUNCTIONS
    :annotation: = {<transfer function index>: <function>, ...}

.. autodata:: XYZ_TO_LINEAR_RGB
    :annotation: = {<color primaries index>: <3x3 matrix>, ...}

.. autodata:: LINEAR_RGB_TO_XYZ
    :annotation: = {<color primaries index>: <3x3 matrix>, ...}


Additional utility functions
----------------------------

The following additional utility functions are provided for the manual
evaluation of certain transform steps.

.. autofunction:: matmul_colors

.. autofunction:: swap_primaries


Color parameter sanity checking
-------------------------------

The :py:func:`sanity_check_video_parameters` function is provided which can
check a given VC-2 video format is 'sane' -- that is it might plausibly be able
to represent some colors.

.. autofunction:: sanity_check_video_parameters

.. autoclass:: ColorParametersSanity
    :members:

"""

import numpy as np

import warnings

from collections import namedtuple

from vc2_data_tables import (
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
    ColorDifferenceSamplingFormats,
)

from vc2_conformance.pseudocode.vc2_math import intlog2


__all__ = [
    "to_xyz",
    "from_xyz",
    "float_to_int_clipped",
    "float_to_int",
    "int_to_float",
    "from_444",
    "to_444",
    "COLOR_MATRICES",
    "INVERSE_COLOR_MATRICES",
    "TRANSFER_FUNCTIONS",
    "INVERSE_TRANSFER_FUNCTIONS",
    "XYZ_TO_LINEAR_RGB",
    "LINEAR_RGB_TO_XYZ",
    "matmul_colors",
    "swap_primaries",
    "ColorParametersSanity",
    "sanity_check_video_parameters",
]


################################################################################
# Linear RGB <--> CIE XYZ Conversion
#
# This conversion step may be implemented using a simple 3x3 matrix
# multiplication. In practice, color primaries are not specified in terms of
# this matrix but instead as a series of chromacities (little-x, little-y) for
# the R, G and B components and the white point. The xy_to_xyz and
# primary_chromacities_to_matrix functions below are used to perform the
# necessary conversions.
################################################################################


PrimaryChromacities = namedtuple(
    "PrimaryChromacities",
    "xw,yw,xr,yr,xg,yg,xb,yb",
)
"""
A specification of a set of color primaries in terms of whitepoint (xw, yx)
and red (xr, yr), green (xg, yg) and blue (xb, yb) chromacities. colors are
specified as :math:`x` and :math:`y` coordinates in the xyY color system.
"""


def xy_to_xyz(x, y):
    """
    Convert a CIE :math:`x` :math:`y` chromacity into XYZ components with unit
    luminance.
    """
    # By definition:
    #
    #    x = X / (X + Y + Z)    and    y = Y / (X + Y + Z)
    #
    # So, by simple rearrangement:

    Y = 1.0
    X = (x / y) * Y
    Z = ((1.0 - x - y) / y) * Y

    return (X, Y, Z)


def primary_chromacities_to_matrix(pc):
    r"""
    Convert a :py:class:`PrimaryChromacities` tuple into a :math:`3 \times 3`
    matrix which converts from linear RGB to XYZ.
    """
    # We want to find a 3x3 matrix m such that:
    #
    #    [ X ]     [       ] [ R ]
    #    [ Y ]  =  [   m   ] [ G ]
    #    [ Z ]     [       ] [ B ]
    #
    # For the supplied set of primaries/whitepoint. We know that for RGB values
    # [1, 0, 0] we get an XYZ value: [Xr, Yr, Zr]:
    #
    #    [ Xr ]     [ xr/yr * Yr         NA NA ] [ 1 ]
    #    [ Yr ]  =  [ Yr                 NA NA ] [ 0 ]
    #    [ Zr ]     [ (1-xr-yr)/yr * Yr  NA NA ] [ 0 ]
    #
    # Due to the definition of x and y w.r.t. XYZ. The middle and rightmost
    # columns are similar but with xg, yg and xb and yb respectively.
    # Consequently we see that m is of the form:
    #
    #    [ X ]     [ xr/yr * Yr         xg/yg * Yg         xb/yb * Yb        ] [ R ]
    #    [ Y ]  =  [ Yr                 Yg                 Yb                ] [ G ]
    #    [ Z ]     [ (1-xr-yr)/yr * Yr  (1-xg-yg)/yg * Yg  (1-xb-yb)/yb * Yb ] [ B ]
    #
    # Lets factor out the currently unknown Y* terms:
    #
    #    [ X ]     [ xr/yr        xg/yg        xb/yb        ] [ Yr  0  0 ] [ R ]
    #    [ Y ]  =  [ 1            1            1            ] [  0 Yg  0 ] [ G ]
    #    [ Z ]     [ (1-xr-yr)/yr (1-xg-yg)/yg (1-xb-yb)/yb ] [  0  0 Yb ] [ B ]
    #
    # We'll refer to the remaining matrix as relative_m:
    #
    #    [ X ]     [            ] [ Yr  0  0 ] [ R ]
    #    [ Y ]  =  [ relative_m ] [  0 Yg  0 ] [ G ]
    #    [ Z ]     [            ] [  0  0 Yb ] [ B ]

    relative_m = np.array(
        [xy_to_xyz(pc.xr, pc.yr), xy_to_xyz(pc.xg, pc.yg), xy_to_xyz(pc.xb, pc.yb)]
    ).T

    # Also by definition, the white point with chromacity xw, yw, and XYZ Xw,
    # Yw, Zw is the color produced by RGB [1, 1, 1]:
    #
    #    [ Xw ]     [            ] [ Yr  0  0 ] [ 1 ]
    #    [ Yw ]  =  [ relative_m ] [  0 Yg  0 ] [ 1 ]
    #    [ Zw ]     [            ] [  0  0 Yb ] [ 1 ]
    #
    # Rearranging
    #
    #    [            ]-1 [ Xw ]     [ Yr  0  0 ] [ 1 ]
    #    [ relative_m ]   [ Yw ]  =  [  0 Yg  0 ] [ 1 ]
    #    [            ]   [ Zw ]     [  0  0 Yb ] [ 1 ]
    #
    #    [            ]-1 [ Xw ]     [ Yr ]
    #    [ relative_m ]   [ Yw ]  =  [ Yg ]
    #    [            ]   [ Zw ]     [ Yb ]
    #
    # By defining the luminance of the whitepoint, Yw, to be 1.0, we can
    # find Xw and Zw from xw and yw:
    #
    #    [            ]-1 [ xw/yw        ]     [ Yr ]     [       ]
    #    [ relative_m ]   [ 1            ]  =  [ Yg ]  =  [ scale ]
    #    [            ]   [ (1-xw-yw)/yw ]     [ Yb ]     [       ]
    #
    # This equation now contains no unknowns other than Yr, Yg and Yb which we
    # can now solve for and together refer to as 'scale':

    scale = np.matmul(
        np.linalg.inv(relative_m),
        xy_to_xyz(pc.xw, pc.yw),
    )

    # We can now apply this to relative_m to get the matrix m:

    m = np.matmul(
        relative_m,
        np.diag(scale),
    )

    return m


LINEAR_RGB_TO_XYZ = {
    PresetColorPrimaries.hdtv:
    # (ITU-R BT.709)
    primary_chromacities_to_matrix(
        PrimaryChromacities(
            xw=0.3127,
            yw=0.3290,
            xr=0.640,
            yr=0.330,
            xg=0.300,
            yg=0.600,
            xb=0.150,
            yb=0.060,
        )
    ),
    PresetColorPrimaries.sdtv_525:
    # (ITU-R BT.601)
    primary_chromacities_to_matrix(
        PrimaryChromacities(
            xw=0.3127,
            yw=0.3290,
            xr=0.630,
            yr=0.340,
            xg=0.310,
            yg=0.595,
            xb=0.155,
            yb=0.070,
        )
    ),
    PresetColorPrimaries.sdtv_625:
    # (ITU-R BT.601)
    primary_chromacities_to_matrix(
        PrimaryChromacities(
            xw=0.3127,
            yw=0.3290,
            xr=0.640,
            yr=0.330,
            xg=0.290,
            yg=0.600,
            xb=0.150,
            yb=0.060,
        )
    ),
    PresetColorPrimaries.d_cinema:
    # (SMPTE ST 428-1)
    np.eye(3),
    PresetColorPrimaries.uhdtv:
    # (ITU-R BT.2020)
    primary_chromacities_to_matrix(
        PrimaryChromacities(
            xw=0.3127,
            yw=0.3290,
            xr=0.708,
            yr=0.292,
            xg=0.170,
            yg=0.797,
            xb=0.131,
            yb=0.046,
        )
    ),
}
r"""
For each set of color primaries in
:py:class:`~vc2_data_tables.PresetColorPrimaries`, a :math:`3 \times 3` matrix
which converts from linear RGB into CIE XYZ.
"""

XYZ_TO_LINEAR_RGB = {key: np.linalg.inv(m) for key, m in LINEAR_RGB_TO_XYZ.items()}
r"""
For each set of color primaries in
:py:class:`~vc2_data_tables.PresetColorPrimaries`, a :math:`3 \times 3` matrix
which converts from CIE XYZ into linear RGB.
"""

################################################################################
# Transfer functions
#
# The following functions implement the transfer functions (gamma functions)
# supported by VC-2. In all cases, these functions are designed such that they
# may be passed either a single value or a numpy array in which an element-wise
# conversion will be performed.
################################################################################


def tv_gamma_transfer_function(l):
    """
    The transfer function from (ITU-R BT.2020).
    """
    alpha = 1.09929682680944
    beta = 0.018053968510807

    # NB: In case of undershoot (l < 0), np.power produces a warning here. The
    # effected values will be overwritten later however so this can be ignored.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        e = alpha * np.power(l, 0.45) - (alpha - 1)
    e = np.where(l < beta, 4.5 * l, e)

    return e


def tv_gamma_inverse_transfer_function(e):
    """
    The reverse of the transfer function from (ITU-R BT.2020).
    """
    alpha = 1.09929682680944
    beta = 0.018053968510807

    # NB: In case of undershoot (l < 0), np.power produces a warning here. The
    # effected values will be overwritten later however so this can be ignored.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        l = np.power((e + 0.099) / alpha, 1 / 0.45)
    l = np.where(e < tv_gamma_transfer_function(beta), e / 4.5, l)

    return l


def extended_gamut_transfer_function(l):
    """
    The transfer function from (ITU-R BT.1361).
    """
    # NB: In case of undershoot (l < 0), np.power produces a warning here. The
    # effected values will be overwritten later however so this can be ignored.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        e = 1.099 * np.power(l, 0.45) - 0.099

    e = np.where(
        l < 0.018,
        4.5 * l,
        e,
    )

    # NB: In case of undershoot (-4 * l < 0), the effected value will not be
    # used
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        e = np.where(
            l < -0.0045,
            -(1.099 * np.power(-4 * l, 0.45) - 0.099) / 4,
            e,
        )

    return e


def linear_transfer_function(l):
    """
    A linear transfer function (for completeness).
    """
    return l


def d_cinema_transfer_function(l):
    """
    The transfer function from (ST 428-1).

    .. note::

        This transfer function does not map 1.0 to 1.0 as other transfer
        functions do. That is, the headroom above the 'reference luminance' for
        luminance values above 1.0 is included in the range 0.0 to 1.0 in the
        transformed signal.
    """
    peak_luminance = 52.37  # cd/m^2
    reference_luminance = 48.0  # cd/m^2

    # Clamp 'l' at 0 as not defined in case of undershoots
    e = np.power((reference_luminance * np.maximum(l, 0)) / peak_luminance, (1 / 2.6))

    return e


def pq_transfer_function(l):
    """
    The Optical-Eelectical Transfer Function (OETF) for Dolby Perceptual
    Quantizer (PQ) (ITU-R BT.2100-2).
    """
    m1 = 2610.0 / 16384.0
    m2 = 2523.0 / 4096.0 * 128
    c1 = 3424.0 / 4096.0
    c2 = 2413.0 / 4096.0 * 32.0
    c3 = 2392.0 / 4096.0 * 32.0

    # Clamp 'l' at 0 as not defined in case of undershoots
    l_m1 = np.power(np.maximum(l, 0), m1)

    e = np.power((c1 + c2 * l_m1) / (1 + c3 * l_m1), m2)

    return e


def hlg_transfer_function(l):
    """
    The Hybrid Log Gamma (HLG) transfer function (ITU-R BT.2100-2).
    """
    a = 0.17883277
    b = 1 - 4 * a
    c = 0.5 - a * np.log(4 * a)

    # NB: In case of undershoot (l < 0), np.log produces a warning here. The
    # effected values will be overwritten later however so this can be ignored.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        e = a * np.log(12 * l - b) + c

    # Clamp 'l' at 0 as not defined in case of undershoots
    e = np.where(
        l <= 1 / 12.0,
        np.sqrt(3 * np.maximum(l, 0)),
        e,
    )

    return e


TRANSFER_FUNCTIONS = {
    PresetTransferFunctions.tv_gamma: tv_gamma_transfer_function,
    PresetTransferFunctions.extended_gamut: extended_gamut_transfer_function,
    PresetTransferFunctions.linear: linear_transfer_function,
    PresetTransferFunctions.d_cinema: d_cinema_transfer_function,
    PresetTransferFunctions.perceptual_quantizer: pq_transfer_function,
    PresetTransferFunctions.hybrid_log_gamma: hlg_transfer_function,
}
"""
For each set of VC-2's, supported transfer functions, a Numpy implementation of
that function. These functions implement the transform from linear to
non-linear RGB, :math:`E_R E_G E_B`. These functions expect and returns a
single value or Numpy array of values.
"""


INVERSE_TRANSFER_FUNCTIONS = {
    PresetTransferFunctions.tv_gamma: tv_gamma_inverse_transfer_function,
    PresetTransferFunctions.linear: linear_transfer_function,
}
"""
For (a subset of) VC-2's, supported transfer functions, a Numpy implementation
of the inverse function. These functions implement the transform from
non-linear to linear RGB. These functions expect and returns a single value or
Numpy array of values.

.. warning::

    An inverse transfer function is currently only provided for
    :py:data:`~vc2_data_tables.PresetTransferFunctions.tv_gamma` because this
    is all that was required at the time of development.
"""


################################################################################
# RGB <--> Y C1 C2 (Color Matrices)
#
# Color matrices are often specified indirectly by a pair of weights, Kr and
# Kb. These are converted into the equivalent 3x3 matrix by
# kr_kb_to_color_matrix.
################################################################################


def kr_kb_to_color_matrix(kr, kb):
    r"""
    Given a description of a Y' Cb Cr color matrix as a :math:`K_r`,
    :math:`K_b` pair, return a :math:`3 \times 3` :math:`E_R E_G E_B` to Y' Cb
    Cr color matrix.
    """
    # The luma component of a Y' Cb Cr signal, Y', is defined as:
    #
    #     Y' = Kr R' + Kg G' + Kb B'
    #
    # Where:
    #
    #     Kr + Kg + Kb = 1
    #
    # So:

    kg = 1 - kr - kb

    # Conceptually raw color difference signals are then given as:
    #
    #     Cb_conceptual = B' - Y'
    #     Cr_conceptual = R' - Y'
    #
    # However these signals must be weighted to give a unity gain. To do this,
    # lets expand Cb_conceptual:
    #
    #     Cb_conceptual = B' - Y'
    #                   = B' - Kr R' - Kg G' - Kb B'
    #                   = (1 - Kb) B' - Kr R' - Kg G'
    #
    # So, manipulating our equation for the unity of Kr, Kg and Kb we can work
    # out what the weighting should be for the equation above:
    #
    #     Kr + Kg + Kb       = 1
    #     Kr + Kg            = 1 - Kb
    #     Kr + Kg + (1 - Kb) = (1 - Kb) + (1 - Kb)
    #                        = 2(1 - Kb)
    #
    # This tells us that the equation above sums to a gain of 2(1 - Kb) so:
    #
    #     Cb = (B' - Y') / 2(1 - Kb)
    #        = ((1 - Kb) B' - Kr R' - Kg G') / 2(1 - Kb)
    #        = (- (Kr / 2(1-Kb)) R' - (Kg / 2(1-Kb)) G' + 0.5 B')
    #
    # And similarly:
    #
    #     Cr = (R' - Y') / 2(1 - Kr)
    #        = (0.5 R' - (Kg / 2(1-Kr)) G' - (Kb / 2(1-Kr)) B')
    #
    # Given these we can now directly construct the color matrix:

    return np.array(
        [
            [kr, kg, kb],
            [-kr / (2 * (1 - kb)), -kg / (2 * (1 - kb)), 0.5],
            [0.5, -kg / (2 * (1 - kr)), -kb / (2 * (1 - kr))],
        ]
    )


COLOR_MATRICES = {
    # (ITU-R BT.709)
    PresetColorMatrices.hdtv: kr_kb_to_color_matrix(kr=0.2126, kb=0.0722),
    # (ITU-R BT.601)
    PresetColorMatrices.sdtv: kr_kb_to_color_matrix(kr=0.2990, kb=0.1140),
    # (ITU-T H.264)
    PresetColorMatrices.reversible: np.array(
        [[+0.25, +0.50, +0.25], [-0.25, +0.50, -0.25], [+0.50, +0.00, -0.50]]
    ),
    # GBR -> Y C1 C2
    PresetColorMatrices.rgb: np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]]),
    # (ITU-R BT.2020)
    PresetColorMatrices.uhdtv: kr_kb_to_color_matrix(kr=0.2627, kb=0.0593),
}
r"""
For each color matrix supported by VC-2, a :math:`3 \times 3` matrix which
transforms from non-linear RGB (:math:`E_R E_G E_B`) to Y C1 C2.
"""


INVERSE_COLOR_MATRICES = {key: np.linalg.inv(m) for key, m in COLOR_MATRICES.items()}
r"""
For each color matrix supported by VC-2, a :math:`3 \times 3` matrix which
transforms from Y C1 C2 to non-linear RGB (:math:`E_R E_G E_B`).
"""


################################################################################
# integer <--> float conversion (11.4.9)
#
# VC-2, like many video systems, deal entirely in positive integer values which
# nominally represent values in the range 0 to +1 or -0.5 to +0.5. These
# functions perform the conversion between these two representations, the
# latter of which is necessary for the color conversion routines and matrices
# above.
################################################################################


def int_to_float(a, offset, excursion):
    """
    Convert (an array of) integer sample values from integers (with the
    specified offset and excursion) to floating point values nominally in the
    range 0 to +1 or -0.5 to +0.5.
    """
    a = np.array(a, dtype=float)
    return (a - offset) / excursion


def float_to_int(a, offset, excursion):
    """
    Convert (an array of) float sample values in the nominal range 0 to +1 or
    -0.5 to +0.5 to integers (with the specified offset and excursion).

    Values which fall outside the range of the integer representation are
    *not* clipped. See :py:func:`float_to_int_clipped`.
    """
    a = np.array(a, dtype=float)

    # Scale/offset
    a = (a * excursion) + offset

    # Round to integer
    a = np.round(a).astype(int)

    return a


def float_to_int_clipped(a, offset, excursion):
    """
    Convert (an array of) float sample values in the nominal range 0 to +1 or
    -0.5 to +0.5 to integers (with the specified offset and excursion).

    Values which fall outside the range of the integer representation are
    clipped.
    """
    a = float_to_int(a, offset, excursion)
    return np.clip(a, 0, (2 ** (intlog2(excursion + 1))) - 1)


################################################################################
# Chroma subsampling format conversion
#
# The funtions below are provided for quick-and-dirty color subsampling
# purposes. The antialiasing filters used are completely inadequate for real
# applications but are sufficient for the production of simple test pictures.
################################################################################


def from_444(chroma, subsampling):
    """
    Subsample a chroma picture component into the specified
    :py:class:`~vc2_data_tables.ColorDifferenceSamplingFormats`.

    .. warning::

        This function uses an extremely crude low-pass filter during
        downsampling which is likely to produce aliasing artefacts. As such,
        pictures produced by this function should not be used for anything
        where high fidelity is required.
    """

    if subsampling == ColorDifferenceSamplingFormats.color_4_4_4:
        return chroma
    elif subsampling == ColorDifferenceSamplingFormats.color_4_2_2:
        h, w = chroma.shape
        new_chroma = np.empty((h, w // 2), dtype=chroma.dtype)

        new_chroma[:, :] = chroma[:, 0::2]
        new_chroma[:, :] += chroma[:, 1::2]
        new_chroma /= 2.0

        return new_chroma
    elif subsampling == ColorDifferenceSamplingFormats.color_4_2_0:
        h, w = chroma.shape
        new_chroma = np.empty((h // 2, w // 2), dtype=chroma.dtype)

        new_chroma[:, :] = chroma[0::2, 0::2]
        new_chroma[:, :] += chroma[0::2, 1::2]
        new_chroma[:, :] += chroma[1::2, 0::2]
        new_chroma[:, :] += chroma[1::2, 1::2]
        new_chroma /= 4.0

        return new_chroma


def to_444(chroma, subsampling):
    """
    Given a chroma picture subsampled according to the specified
    :py:class:`~vc2_data_tables.ColorDifferenceSamplingFormats`, return an
    upsampled chroma signal.

    .. warning::

        This function uses an extremely crude anti-aliasing filter during
        upsampling which is likely to produce artefacts. As such, pictures
        produced by this function should not be used for anything where high
        fidelity is required.
    """
    if subsampling == ColorDifferenceSamplingFormats.color_4_4_4:
        return chroma
    elif subsampling == ColorDifferenceSamplingFormats.color_4_2_2:
        return np.repeat(chroma, 2, axis=1)
    elif subsampling == ColorDifferenceSamplingFormats.color_4_2_0:
        return np.repeat(np.repeat(chroma, 2, axis=1), 2, axis=0)


################################################################################
# High-level utility functions
################################################################################


def to_xyz(y, c1, c2, video_parameters):
    r"""
    Convert a picture from a native VC-2 integer Y C1 C2 format into floating
    point CIE XYZ format.

    Parameters
    ==========
    y, c1, c2 : :py:class:`numpy.array`
        Three 2D :py:class:`numpy.arrays <numpy.array>` containing integer Y C1
        C2 values for a picture.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The VC-2 parameters describing the video format in use. The following
        fields are required:

        * ``color_diff_format_index``
        * ``luma_offset``
        * ``luma_excursion``
        * ``color_diff_offset``
        * ``color_diff_excursion``
        * ``color_primaries``
        * ``color_matrix``
        * ``transfer_function``

    Returns
    =======
    yxz : :py:class:`numpy.array`
        A 3D :py:class:`numpy.array` with dimensions ``(height, width, 3)``
        containing floating point CIE XYZ values for a picture.
    """
    # Convert each component into floating point, 0 - +1 or -0.5 - +0.5 form
    y_float = int_to_float(
        y,
        video_parameters["luma_offset"],
        video_parameters["luma_excursion"],
    )
    c1_float = int_to_float(
        c1,
        video_parameters["color_diff_offset"],
        video_parameters["color_diff_excursion"],
    )
    c2_float = int_to_float(
        c2,
        video_parameters["color_diff_offset"],
        video_parameters["color_diff_excursion"],
    )

    # Upsample to 4:4:4
    c1_upsampled = to_444(c1_float, video_parameters["color_diff_format_index"])
    c2_upsampled = to_444(c2_float, video_parameters["color_diff_format_index"])

    # Produce a Nx3 matrix with each column containing a Y C1 C2 value.
    yc1c2_cols = np.stack(
        [y_float.reshape(-1), c1_upsampled.reshape(-1), c2_upsampled.reshape(-1)],
        axis=0,
    )

    # Convert to non-linear RGB
    inverse_color_matrix = INVERSE_COLOR_MATRICES[
        video_parameters["color_matrix_index"]
    ]
    eregeb_cols = np.matmul(inverse_color_matrix, yc1c2_cols)

    # Convert to linear RGB
    inverse_transfer_function = INVERSE_TRANSFER_FUNCTIONS[
        video_parameters["transfer_function_index"]
    ]
    linear_rgb_cols = inverse_transfer_function(eregeb_cols)

    # Convert to CIE XYZ
    rgb_to_xyz_matrix = LINEAR_RGB_TO_XYZ[video_parameters["color_primaries_index"]]
    xyz_cols = np.matmul(rgb_to_xyz_matrix, linear_rgb_cols)

    # Convert to 3D array of XYZ values
    xyz = xyz_cols.T.reshape(y.shape + (3,))

    return xyz


def from_xyz(xyz, video_parameters):
    """
    Convert a picture from CIE XYZ format into a native VC-2 integer, chroma
    subsampled Y C1 C2 format.

    Parameters
    ==========
    yxz : :py:class:`numpy.array`
        A 3D :py:class:`numpy.array` with dimensions ``(height, width, 3)``
        containing floating point CIE XYZ values for a picture.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
        The VC-2 parameters describing the video format to produce. The following
        fields are required:

        * ``color_diff_format_index``
        * ``luma_offset``
        * ``luma_excursion``
        * ``color_diff_offset``
        * ``color_diff_excursion``
        * ``color_primaries``
        * ``color_matrix``
        * ``transfer_function``

    Returns
    =======
    y, c1, c2 : :py:class:`numpy.array`
        A set of three 2D :py:class:`numpy.array` containing integer Y C1 C2
        values for a picture. If chroma subsampling is used, the C1 and C2
        arrays may differ in size from the Y component.
    """
    # Convert to columns of XYZ values
    xyz_cols = xyz.reshape(-1, 3).T

    # Convert to linear RGB
    xyz_to_rgb_matrix = XYZ_TO_LINEAR_RGB[video_parameters["color_primaries_index"]]
    linear_rgb_cols = np.matmul(xyz_to_rgb_matrix, xyz_cols)

    # Convert to non-linear RGB
    transfer_function = TRANSFER_FUNCTIONS[video_parameters["transfer_function_index"]]
    eregeb_cols = transfer_function(linear_rgb_cols)

    # Convert to Y C1 C2
    color_matrix = COLOR_MATRICES[video_parameters["color_matrix_index"]]
    yc1c2_cols = np.matmul(color_matrix, eregeb_cols)

    # Convert to 3D
    yc1c2 = yc1c2_cols.T.reshape(xyz.shape)

    # Color subsample
    y = yc1c2[:, :, 0]
    c1 = yc1c2[:, :, 1]
    c2 = yc1c2[:, :, 2]

    c1_subsampled = from_444(c1, video_parameters["color_diff_format_index"])
    c2_subsampled = from_444(c2, video_parameters["color_diff_format_index"])

    # Convert to integer ranges
    y_int = float_to_int_clipped(
        y,
        video_parameters["luma_offset"],
        video_parameters["luma_excursion"],
    )
    c1_int = float_to_int_clipped(
        c1_subsampled,
        video_parameters["color_diff_offset"],
        video_parameters["color_diff_excursion"],
    )
    c2_int = float_to_int_clipped(
        c2_subsampled,
        video_parameters["color_diff_offset"],
        video_parameters["color_diff_excursion"],
    )

    return (y_int, c1_int, c2_int)


def matmul_colors(matrix, array):
    r"""
    Given a (height, width, 3) 3D array, return a new 3D array where each
    triple in the first array has been multiplied by the specified :math:`3
    \times 3` matrix.
    """
    return np.matmul(
        matrix,
        array.reshape(-1, 3).T,
    ).T.reshape(array.shape)


def swap_primaries(xyz, video_parameters_before, video_parameters_after):
    r"""
    Given an image defined in terms of one set of primaries, return a new image
    defined in terms of a different set of primaries but with the same
    numerical R, G and B values under the new set of primaries.

    This transformation is useful when an image is defined not by absolute
    colors but rather colors relative to whatever primaries are in use. For
    example, a test pattern designed to show swatches of pure color primaries
    may be given relative to a particular set of primaries but needs to be
    adapted for use with another set of primaries.

    Parameters
    ==========
    xyz : :math:`3 \times 3` array (height, width, 3)
    video_parameters_before : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    video_parameters_after : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`

    Returns
    =======
    xyz : :math:`3 \times 3` array (height, width, 3)
    """  # noqa: E501
    xyz_to_linear_rgb_before = XYZ_TO_LINEAR_RGB[
        video_parameters_before["color_primaries_index"]
    ]
    linear_rgb_after_to_xyz = LINEAR_RGB_TO_XYZ[
        video_parameters_after["color_primaries_index"]
    ]

    m = np.matmul(linear_rgb_after_to_xyz, xyz_to_linear_rgb_before)

    return matmul_colors(m, xyz)


class ColorParametersSanity(object):
    """
    Result of :py:func:`sanity_check_video_parameters`. Indicates the sanity
    (or insanity) of a set of video parameters.

    Truthy if sane, falsey otherwise.

    Use the various properties to determine what is and is not sane.

    Use the :py:meth:`explain` function to return a string with a human
    readable explanation.
    """

    def __init__(
        self,
        luma_depth_sane=True,
        color_diff_depth_sane=True,
        black_sane=True,
        white_sane=True,
        red_sane=True,
        green_sane=True,
        blue_sane=True,
        color_diff_format_sane=True,
        luma_vs_color_diff_depths_sane=True,
    ):
        self._luma_depth_sane = luma_depth_sane
        self._color_diff_depth_sane = color_diff_depth_sane
        self._black_sane = black_sane
        self._white_sane = white_sane
        self._red_sane = red_sane
        self._green_sane = green_sane
        self._blue_sane = blue_sane
        self._color_diff_format_sane = color_diff_format_sane
        self._luma_vs_color_diff_depths_sane = luma_vs_color_diff_depths_sane

    @property
    def luma_depth_sane(self):
        """
        True iff the luma component has been assigned at least 8 bits.
        """
        return self._luma_depth_sane

    @property
    def color_diff_depth_sane(self):
        """
        True iff the color difference components have been assigned at least 8
        bits.
        """
        return self._color_diff_depth_sane

    @property
    def black_sane(self):
        """
        If True, the format can represent (video) black.
        """
        return self._black_sane

    @property
    def white_sane(self):
        """
        If True, the format can represent (video) white.
        """
        return self._white_sane

    @property
    def red_sane(self):
        """
        If True, the format can represent (video) primary red.
        """
        return self._red_sane

    @property
    def green_sane(self):
        """
        If True, the format can represent (video) primary green.
        """
        return self._green_sane

    @property
    def blue_sane(self):
        """
        If True, the format can represent (video) primary blue.
        """
        return self._blue_sane

    @property
    def color_diff_format_sane(self):
        """
        True iff the color difference sampling format is appropriate for the
        color format.

        False when non-4:4:4 sampling is used for RGB formats.
        """
        return self._color_diff_format_sane

    @property
    def luma_vs_color_diff_depths_sane(self):
        """
        True iff the relative offsets/excursions of luma and color difference
        components are appropriately matched.

        False when not identical for RGB formats.
        """
        return self._luma_vs_color_diff_depths_sane

    def __bool__(self):
        """
        True iff all sanity checks have come up sane.
        """
        return all(
            (
                self.luma_depth_sane,
                self.color_diff_depth_sane,
                self.black_sane,
                self.white_sane,
                self.red_sane,
                self.green_sane,
                self.blue_sane,
                self.color_diff_format_sane,
                self.luma_vs_color_diff_depths_sane,
            )
        )

    def __nonzero__(self):
        """Provided for Python 2.x backward compatibility."""
        return self.__bool__()

    def explain(self):
        """
        Return a human-readable explanation of why a video format is not sane
        (or simply state that it is sane, if it is).
        """
        out = ""

        insane_depths = []
        if not self.luma_depth_sane:
            insane_depths.append("luma")
        if not self.color_diff_depth_sane:
            insane_depths.append("color_diff")

        if insane_depths:
            out += (
                "The {} value{} indicate{} a video format "
                "using fewer than 8 bits. Hint: The *_excursion field "
                "gives the range of legal values for a component (i.e. "
                "max_value - min_value), not a number of bits.\n\n"
            ).format(
                " and ".join(
                    "{}_excursion".format(component) for component in insane_depths
                ),
                "s" if len(insane_depths) != 1 else "",
                "s" if len(insane_depths) == 1 else "",
            )

        insane_colors = [
            color
            for color in ["black", "white", "red", "green", "blue"]
            if not getattr(self, "{}_sane".format(color))
        ]

        if insane_colors:
            out += (
                "Some colors (e.g. {}) cannot be represented "
                "in the video format specified. "
                "Hint: Check luma_offset is a near zero value, "
                "for Y C1 C2 formats check color_diff_offset is near the "
                "center of the signal range "
                "and for RGB formats it is near zero.\n\n"
            ).format(", ".join(insane_colors))

        if not self.color_diff_format_sane:
            out += (
                "A color subsampling format other than 4:4:4 has been used "
                "for format using the RGB color matrix.\n\n"
            )

        if not self.luma_vs_color_diff_depths_sane:
            out += (
                "Different (luma|color_diff)_offset and/or "
                "(luma|color_diff)_excursion values have been specified "
                "for format using the RGB color matrix.\n\n"
            )

        if out == "":
            return "Color format is sensible."
        else:
            return out.strip()

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            ", ".join(
                "{}=False".format(name)
                for name in dir(self)
                if (
                    not name.startswith("_")
                    and name.endswith("_sane")
                    and getattr(self, name) is False
                )
            ),
        )


def sanity_check_video_parameters(video_parameters):
    r"""
    Given a set of
    :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`, check that a
    set of video parameters could plausibly be used to encode a color signal
    (regardless of whether the color specification itself is sensible).

    Specifically, the following checks are carried out:

    * Are the luma and color difference signals at least 8 bits?
    * Can white, black and saturated primary red, green and blue be encoded?
    * When the RGB color matrix is used:
        * Is the color difference sampling mode 4:4:4?
        * Are the luma and chroma components the same depth?

    Returns a :py:class:`ColorParametersSanity` as a result.
    """
    # Check at least 8 bits
    luma_depth_sane = video_parameters["luma_excursion"] >= 128
    color_diff_depth_sane = video_parameters["color_diff_excursion"] >= 128

    # Create 5 columns with black/white & RGB primaries as RGB values
    linear_rgb_cols = np.array(
        [
            # Blk  Wht  Red  Grn  Blu
            [0.0, 1.0, 1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 1.0, 0.0],
            [0.0, 1.0, 0.0, 0.0, 1.0],
        ]
    )

    # Convert to non-linear RGB
    transfer_function = TRANSFER_FUNCTIONS[video_parameters["transfer_function_index"]]
    eregeb_cols = transfer_function(linear_rgb_cols)

    # Convert to Y C1 C2
    color_matrix = COLOR_MATRICES[video_parameters["color_matrix_index"]]
    yc1c2_cols = np.matmul(color_matrix, eregeb_cols)

    # Check if the integer ranges overflow (NB: we still allow for tiny
    # overflows due to rounding errors)
    clipped_cols = np.all(
        [
            abs(
                float_to_int(array, offset, excursion)
                - float_to_int_clipped(array, offset, excursion)
            )
            <= 1
            for array, offset, excursion in [
                (
                    yc1c2_cols[0],
                    video_parameters["luma_offset"],
                    video_parameters["luma_excursion"],
                ),
                (
                    yc1c2_cols[1],
                    video_parameters["color_diff_offset"],
                    video_parameters["color_diff_excursion"],
                ),
                (
                    yc1c2_cols[2],
                    video_parameters["color_diff_offset"],
                    video_parameters["color_diff_excursion"],
                ),
            ]
        ],
        axis=0,
    )

    black_sane = clipped_cols[0]
    white_sane = clipped_cols[1]
    red_sane = clipped_cols[2]
    green_sane = clipped_cols[3]
    blue_sane = clipped_cols[4]

    # Check special format considerations for RGB signals
    if video_parameters["color_matrix_index"] == PresetColorMatrices.rgb:
        color_diff_format_sane = (
            video_parameters["color_diff_format_index"]
            == ColorDifferenceSamplingFormats.color_4_4_4
        )

        luma_vs_color_diff_depths_sane = video_parameters[
            "luma_offset"
        ] == video_parameters["color_diff_offset"] and (
            video_parameters["luma_excursion"]
            == video_parameters["color_diff_excursion"]
        )
    else:
        color_diff_format_sane = True
        luma_vs_color_diff_depths_sane = True

    return ColorParametersSanity(
        luma_depth_sane=luma_depth_sane,
        color_diff_depth_sane=color_diff_depth_sane,
        black_sane=black_sane,
        white_sane=white_sane,
        red_sane=red_sane,
        green_sane=green_sane,
        blue_sane=blue_sane,
        color_diff_format_sane=color_diff_format_sane,
        luma_vs_color_diff_depths_sane=luma_vs_color_diff_depths_sane,
    )



from vc2_data_tables import PictureCodingModes

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.encoder import make_sequence

from vc2_conformance.picture_generators import (
    mid_gray,
    repeat_pictures,
)


@decoder_test_case_generator
def picture_numbers(codec_features):
    """
    **Tests picture numbers are correctly read from the bitstream.**

    Each test case contains 8 blank pictures numbered in a particular way.

    ``picture_numbers[start_at_zero]``
        The first picture has picture number 0.

    ``picture_numbers[non_zero_start]``
        The first picture has picture number 1000.

    ``picture_numbers[wrap_around]``
        The first picture has picture number 4294967292, with the picture
        numbers wrapping around to 0 on the 4th picture in the sequence.

    ``picture_numbers[odd_first_picture]``
        The first picture has picture number 7. This test case is only included
        when the picture coding mode is 0 (i.e. pictures are frames) since the
        first field of each frame must have an even number when the picture
        coding mode is 1 (i.e. pictures are fields) (11.5).
    """
    # Create a sequence with at least 8 pictures (and 4 frames)
    mid_gray_pictures = list(
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        )
    )
    mid_gray_pictures = list(
        repeat_pictures(
            mid_gray_pictures,
            8 // len(mid_gray_pictures),
        )
    )

    test_cases = [
        ("start_at_zero", [0, 1, 2, 3, 4, 5, 6, 7]),
        ("non_zero_start", [1000, 1001, 1002, 1003, 1004, 1005, 1006, 1007]),
        ("wrap_around", [4294967292, 4294967293, 4294967294, 4294967295, 0, 1, 2, 3]),
    ]

    if codec_features["picture_coding_mode"] == PictureCodingModes.pictures_are_frames:
        test_cases.append(("odd_first_picture", [7, 8, 9, 10, 11, 12, 13, 14]))

    for description, picture_numbers in test_cases:
        yield TestCase(
            Stream(
                sequences=[
                    make_sequence(
                        codec_features,
                        [
                            dict(picture, pic_num=pic_num)
                            for picture, pic_num in zip(
                                mid_gray_pictures, picture_numbers
                            )
                        ],
                    )
                ]
            ),
            description,
        )



"""
Tests which verify that codecs correctly ignore padding data units (10.4.5).
"""

from copy import deepcopy

from vc2_data_tables import ParseCodes

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.picture_generators import mid_gray, repeat_pictures

from vc2_conformance.encoder import (
    make_sequence,
    IncompatibleLevelAndDataUnitError,
)

from vc2_conformance.test_cases.decoder.common import make_dummy_end_of_sequence


def replace_padding_data(sequence, bytes):
    """
    Replace the bytes in all :py:class:`~vc2_conformance.bitstream.Padding`
    data units in a :py:class:`~vc2_conformance.bitstream.Sequence`.

    A new sequence is returned and the old sequence left unmodified.
    """
    sequence = deepcopy(sequence)
    for data_unit in sequence["data_units"]:
        if data_unit["parse_info"]["parse_code"] == ParseCodes.padding_data:
            data_unit["padding"]["bytes"] = bytes

    return sequence


@decoder_test_case_generator
def padding_data(codec_features):
    """
    **Tests that the contents of padding data units are ignored.**

    This test case consists of a sequence containing two blank frames in which
    every-other data unit is a padding data unit (10.4.5) of various lengths
    and contents (described below).

    ``padding_data[empty]``
        Padding data units containing zero padding bytes (i.e. just consisting
        of a parse info header).

    ``padding_data[zero]``
        Padding data units containing 32 bytes set to 0x00.

    ``padding_data[non_zero]``
        Padding data units containing 32 bytes containing the ASCII encoding of
        the text ``Ignore this padding data please!``.

    ``padding_data[dummy_end_of_sequence]``
        Padding data units containing 32 bytes containing an encoding of an end
        of sequence data unit (10.4.1).

    Where padding data units are not permitted by the VC-2 level in use, these
    test cases are omitted.
    """
    # Generate a base sequence in which we'll modify the padding data units. We
    # ensure there are always at least two pictures in the sequences to make
    # premature termination obvious.
    try:
        base_sequence = make_sequence(
            codec_features,
            repeat_pictures(
                mid_gray(
                    codec_features["video_parameters"],
                    codec_features["picture_coding_mode"],
                ),
                2,
            ),
            # Insert padding data between every data unit
            "sequence_header (padding_data .)* padding_data end_of_sequence $",
        )
    except IncompatibleLevelAndDataUnitError:
        # Padding not allowed in the supplied video format so just skip this
        # test
        return

    for description, data in [
        (
            "empty",
            b"",
        ),
        (
            "zero",
            b"\x00" * 32,
        ),
        (
            "non_zero",
            b"Ignore this padding data please!",
        ),
        (
            "dummy_end_of_sequence",
            make_dummy_end_of_sequence().ljust(32, b"\x00"),
        ),
    ]:
        yield TestCase(
            Stream(sequences=[replace_padding_data(base_sequence, data)]),
            description,
        )



"""
The :py:mod:`vc2_conformance.picture_generators` module contains routines for
generating video sequences for arbitrary VC-2 video formats.

The picture generators in this module are used to generate encoder and decoder
test cases (:py:mod:`vc2_conformance.test_cases`).

All picture generator functions take a
:py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`, and
:py:class:`~vc2_data_tables.PictureCodingModes` as their arguments, defining the
desired VC-2 video format. They then yield a sequence of ``{"Y": [[int, ...],
...], "C1": [[int, ...], ...], "C2": [[int, ...], ...], "pic_num": int}``
dictionaries containing integer picture component values, ready for encoding or
writing to a file.

The following picture generators are provided.

.. autofunction:: moving_sprite

.. autofunction:: static_sprite

.. autofunction:: mid_gray

.. autofunction:: white_noise

.. autofunction:: linear_ramps

Longer sequences may be produced by repeating the pictures produced by the
above generators:

.. autofunction:: repeat_pictures

.. note::

    Where the picture coding mode indicates that pictures represents fields
    (not frames), an even number of pictures (i.e. a whole number of frames) is
    always generated. This means that, for example, a single-frame picture
    generator will generate *two* pictures for these formats.

"""

import functools

from copy import deepcopy

from itertools import cycle

from fractions import Fraction

import math

import numpy as np

from PIL import Image

from vc2_conformance.file_format import read

from vc2_conformance.dimensions_and_depths import compute_dimensions_and_depths

from vc2_data_tables import (
    PictureCodingModes,
    SourceSamplingModes,
)

from vc2_conformance.color_conversion import (
    LINEAR_RGB_TO_XYZ,
    to_xyz,
    from_xyz,
    matmul_colors,
    swap_primaries,
)

from vc2_conformance_data import (
    POINTER_SPRITE_FILENAME,
    NATURAL_PICTURES_FILENAMES,
)


__all__ = [
    "moving_sprite",
    "static_sprite",
    "mid_gray",
    "white_noise",
    "linear_ramps",
    "repeat_pictures",
]


def read_as_xyz(filename):
    """
    Read a VC-2 raw image (see :py:mod:`vc2_conformance.file_format`) into a
    floating point CIE XYZ color 3D array.

    Returns
    =======
    xyz : :py:class:`numpy.array`
        A (height, width, 3) array containing the loaded picture, upsampled to
        4:4:4 color subsampling and converted to CIE XYZ color values.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    """
    picture, video_parameters, picture_coding_mode = read(filename)

    xyz = to_xyz(
        np.array(picture["Y"]),
        np.array(picture["C1"]),
        np.array(picture["C2"]),
        video_parameters,
    )

    return xyz, video_parameters, picture_coding_mode


def resize(im, width, height):
    """
    Resize a linear-light 3D image.
    """
    # Here we use PIL's resize function with a reasonably high-quality LANCZOS
    # filter. Since PIL only supports single-channel floating point images, we
    # must process each channel separately.
    return np.stack(
        [
            Image.fromarray(im[:, :, channel]).resize((width, height), Image.LANCZOS)
            for channel in range(3)
        ],
        axis=-1,
    )


def read_as_xyz_to_fit(filename, width, height, pixel_aspect_ratio=1):
    """
    Read a VC-2 raw image (see :py:mod:`vc2_conformance.file_format`) into a
    floating point CIE XYZ color 3D array, sized to fit the desired picture
    size, aspect ratio and pixel aspect ratio.

    Parameters
    ==========
    filename : str
        Filename to read from.
    width : int
    height : int
        Width and height of the desired output image.
    pixel_aspect_ratio : :py:class:`Fraction`
        The pixel aspect ratio of the output image.
    """
    image, image_video_parameters, image_picture_coding_mode = read_as_xyz(filename)
    assert image_video_parameters["pixel_aspect_ratio_numer"] == 1
    assert image_video_parameters["pixel_aspect_ratio_denom"] == 1

    im_height, im_width, _ = image.shape

    # Stretch original image to correct for the target pixel aspect ratio
    if pixel_aspect_ratio.numerator != pixel_aspect_ratio.denominator:
        im_width *= pixel_aspect_ratio.denominator
        im_width //= pixel_aspect_ratio.numerator

    im_aspect = Fraction(im_width, im_height)
    out_aspect = Fraction(width, height)

    if im_aspect > out_aspect:
        # Image wider than output, squash so height fits then crop width
        im_height = height
        im_width = int(math.ceil(im_height * im_aspect))
    else:
        # Image taller than output, squash so width fits then crop height
        im_width = width
        im_height = int(math.ceil(im_width / im_aspect))

    image = resize(image, im_width, im_height)

    # Crop to output shape
    x_excess = im_width - width
    y_excess = im_height - height
    return image[
        y_excess // 2 : (y_excess // 2) + height,
        x_excess // 2 : (x_excess // 2) + width,
        :,
    ]


def frames_to_samples(video_parameters, frames):
    """
    Convert a number of seconds into a number of samples (frames for
    progressive video, fields for interlaced video).

    Returns a (num_samples, relative_rate) pair. The 'relative_rate' value
    gives the relative sample rate to the frame rate and is 1 for progressive
    and 2 for interlaced formats.
    """
    if video_parameters["source_sampling"] == SourceSamplingModes.interlaced:
        relative_rate = 2
    else:
        relative_rate = 1

    num_samples = frames * relative_rate

    return (num_samples, relative_rate)


def progressive_to_interlaced(video_parameters, picture_coding_mode, pictures):
    """
    Given a sequence of frames, produces a sequence of half-height fields.
    There is a 1:1 mapping between incoming progressive frames and outgoing
    interlaced fields with half of the incoming lines being dropped from each
    frame.
    """
    first_row_indices = [0, 1] if video_parameters["top_field_first"] else [1, 0]

    for first_row, picture in zip(cycle(first_row_indices), pictures):
        yield picture[first_row::2, :, :]


def progressive_to_split_fields(video_parameters, picture_coding_mode, pictures):
    """
    Given a sequence of frames, produces a series of half-height fields. Each
    frame is turned into a pair of successive fields.
    """
    first_row_indices = [0, 1] if video_parameters["top_field_first"] else [1, 0]

    for picture in pictures:
        for first_row in first_row_indices:
            yield picture[first_row::2, :, :]


def interleave_fields(video_parameters, picture_coding_mode, pictures):
    """
    Given a sequence of fields, interleave those fields into frames. Each
    successive pair of fields will be interleaved into a single output frame.
    """
    it = iter(pictures)
    for field_pair in zip(it, it):
        if not video_parameters["top_field_first"]:
            field_pair = field_pair[::-1]

        top, bottom = field_pair

        interleaved = np.empty((top.shape[0] * 2, top.shape[1], 3), dtype=top.dtype)
        interleaved[0::2, :, :] = top
        interleaved[1::2, :, :] = bottom

        yield interleaved


def progressive_to_pictures(video_parameters, picture_coding_mode, pictures):
    """
    Given a sequence of frames, produce a series of pictures for the current
    source sampling and picture coding modes.

    When interlacing is used, successive pictures will be vertically subsampled
    effectively reducing the frame rate by half.
    """
    pictures_are_frames = picture_coding_mode == PictureCodingModes.pictures_are_frames
    progressive = video_parameters["source_sampling"] == SourceSamplingModes.progressive

    if pictures_are_frames:
        if progressive:
            pass  # Nothing to do
        else:  # interlaced
            # Interlace and combine two fields to a picture
            pictures = progressive_to_interlaced(
                video_parameters, picture_coding_mode, pictures
            )
            pictures = interleave_fields(
                video_parameters, picture_coding_mode, pictures
            )
    else:  # pictures are fields
        if progressive:
            # Split each frame across two pictures
            pictures = progressive_to_split_fields(
                video_parameters, picture_coding_mode, pictures
            )
        else:  # interlaced
            # Interlaced, one field per picture
            pictures = progressive_to_interlaced(
                video_parameters, picture_coding_mode, pictures
            )

    return pictures


def xyz_to_native(video_parameters, picture_coding_mode, pictures):
    """
    Given a sequence of CIE XYZ pictures as 3D arrays, produce a corresponding
    sequence of dictionaries for the specified video format of the form::

        {
            "Y": [[int, ...], ...],
            "C1": [[int, ...], ...],
            "C2": [[int, ...], ...],
            "pic_num": int,
        }

    As constructed by the VC-2 pseudocode.

    Picture numbers are assigned starting from zero.
    """
    for pic_num, picture in enumerate(pictures):
        y, c1, c2 = from_xyz(picture, video_parameters)
        yield {
            "Y": y.tolist(),
            "C1": c1.tolist(),
            "C2": c2.tolist(),
            "pic_num": pic_num,
        }


def pipe(next_function):
    """
    Turn a function of the form::

        f(video_parameters, iterable) -> iterable

    Into a decorator for processing the outputs of video sequence generator
    functions of the form::

        g(video_parameters, ...) -> iterable

    For example::

        >>> def repeat_pictures(video_parameters, picture_coding_mode, iterable):
        ...     for picture in iterable:
        ...         yield picture
        ...         yield picture

        >>> @pipe(repeat_pictures)
        ... def picture_generator(video_parameters, picture_coding_mode, num_frames):
        ...     for frame in range(num_frames):
        ...         picture = something(...)
        ...         yield picture

    In this example, the ``picture_generator`` generator function will generate a
    series of repeated pictures.

    Note that the first argument (``video_parameters``) to the decorated
    function is also passed to ``next_function``.
    """

    @functools.wraps(next_function)
    def decorator(video_sequence_generator):
        @functools.wraps(video_sequence_generator)
        def wrapper(video_parameters, picture_coding_mode, *args, **kwargs):
            iterator = video_sequence_generator(
                video_parameters, picture_coding_mode, *args, **kwargs
            )
            return next_function(video_parameters, picture_coding_mode, iterator)

        return wrapper

    return decorator


def read_and_adapt_pointer_sprite(video_parameters):
    """
    Read the :py:data:`vc2_conformance_data.POINTER_SPRITE_FILENAME` sprite,
    correcting to match the pixel aspect ratio and color primaries specified in
    the provided ``video_parameters``.

    Returns a CIE-XYZ image.
    """
    sprite, sprite_video_parameters, _ = read_as_xyz(POINTER_SPRITE_FILENAME)

    # This sprite has been designed with the intention that 'white', 'red',
    # 'green', and 'blue' in the sprite's color model should be transformed
    # into the native primaries of the output format. As a consequence, we swap
    # the color primaries here.
    sprite = swap_primaries(sprite, sprite_video_parameters, video_parameters)

    sprite_width = sprite_video_parameters["frame_width"]
    sprite_height = sprite_video_parameters["frame_height"]
    assert sprite_video_parameters["pixel_aspect_ratio_numer"] == 1
    assert sprite_video_parameters["pixel_aspect_ratio_denom"] == 1

    # Make sprite square under all pixel aspect ratios
    if (
        video_parameters["pixel_aspect_ratio_numer"]
        != video_parameters["pixel_aspect_ratio_denom"]
    ):
        sprite_width *= video_parameters["pixel_aspect_ratio_denom"]
        sprite_width //= video_parameters["pixel_aspect_ratio_numer"]
        sprite = resize(sprite, sprite_width, sprite_height)

    return sprite


def repeat_pictures(pictures, count):
    """
    Repeat a sequence of pictures produced by a picture generator to produce a
    longer sequence.
    """
    for pic_num, picture in enumerate(list(pictures) * count):
        picture = deepcopy(picture)
        picture["pic_num"] = pic_num
        yield picture


@pipe(xyz_to_native)
@pipe(progressive_to_pictures)
def real_pictures(video_parameters, picture_coding_mode):
    """
    A video sequence containing a series of real pictures.

    This image sequence is provided as an easy visual sanity checking measure
    for decoders and a basic performance check for encoders.
    """
    for filename in NATURAL_PICTURES_FILENAMES:
        picture = read_as_xyz_to_fit(
            filename,
            video_parameters["frame_width"],
            video_parameters["frame_height"],
            Fraction(
                video_parameters["pixel_aspect_ratio_numer"],
                video_parameters["pixel_aspect_ratio_denom"],
            ),
        )

        # NB: Repeat pictures for interlaced modes so each frame is made up of
        # two fields containing the same picture content
        yield picture
        if video_parameters["source_sampling"] == SourceSamplingModes.interlaced:
            yield picture


@pipe(xyz_to_native)
@pipe(progressive_to_pictures)
def moving_sprite(video_parameters, picture_coding_mode, num_frames=10):
    """
    A video sequence containing a simple moving synthetic image.

    This sequence consists of a 128 by 128 pixel sprite (shown below) on a
    black background which traverses the screen from left-to-right moving 16
    pixels to the right every frame (or 8 every field). By default the sequence
    is 10 frames long.

    .. image:: /_static/test_images/pointer.png

    This test sequence may be used to verify that interlacing, pixel aspect
    ratio and frame-rate metadata is being correctly reported by a codec for
    display purposes.

    For interlaced formats (see ``scan_format`` (11.4.5)), sequential fields
    will contain the sprite at different horizontal positions, regardless of
    whether pictures are fields or frames (see picture coding mode (11.5)). As
    a result, when frames are viewed as a set of raw interleaved fields, ragged
    edges will be visible.

    Conversely, for progressive formats, sequential fields contain alternate
    lines from the same moment in time and when interleaved should produce
    smooth edges, regardless of the picture coding mode.

    In the very first field of the sequence, the left edge of the white
    triangle will touch the edge of the frame.  In interlaced formats, the top
    line of white pixels in the sprite will always be located on the top field.
    As a result, the line immediately below should always appear shifted to the
    right when top-field-first field order is used and shifted to the left when
    bottom-field-first order is used (see 'top field first parameter' (11.3)).

    The sprite should be square with the white triangle having equal height and
    length and the hypotenuse lying at an angle of 45 degrees. The circular
    cut-out should be a perfect circle. This verifies that the pictures are
    displayed with the correct pixel aspect ratio (11.4.7).

    The text in the sprite is provided to check that the correct picture
    orientation has been used.

    The colors of the characters 'V', 'C' and '2' are colored saturated
    primary red, green, and blue for the color primaries used (11.4.10.2). This
    provides a basic verification that the color components have been provided
    to the display system and decoded correctly.
    """
    sprite = read_and_adapt_pointer_sprite(video_parameters)

    frame_width = video_parameters["frame_width"]
    frame_height = video_parameters["frame_height"]
    sprite_height, sprite_width, _ = sprite.shape

    # Generate frames
    num_samples, relative_rate = frames_to_samples(video_parameters, num_frames)
    x_step_size = 16 // relative_rate
    for px in np.arange(num_samples, dtype=np.int64) * x_step_size:
        # For bizarre, tiny picture formats too small for the sprite, clip the
        # sprite to fit
        sw = min(frame_width, sprite_width)
        sh = min(frame_height, sprite_height)

        # Clip sprite at edge of screen
        sx = 0
        if px < 0:
            sw += px
            sx -= px
            px -= px

        # Clip sprite at right edge of screen
        if px + sw > frame_width:
            sw -= (px + sw) - frame_width
            # Special case: picture is off edge of display. Sould only occur
            # with absurdly small (e.g. < 8 px wide) frame sizes.
            sw = max(0, sw)

        # Blit the sprite onto an otherwise empty frame
        picture = np.zeros((frame_height, frame_width, 3))
        picture[:sh, px : px + sw, :] = sprite[:sh, sx : sx + sw, :]

        yield picture


@pipe(xyz_to_native)
@pipe(progressive_to_pictures)
def static_sprite(video_parameters, picture_coding_mode):
    """
    A video sequence containing a exactly one frame containing a synthetic
    image.

    This sequence consists of a 128 by 128 pixel sprite (shown below) located
    at the top-left corner of the frame on a black background.

    .. image:: /_static/test_images/pointer.png

    This test sequence may be used to verify that interlacing, pixel aspect
    ratio and frame-rate metadata is being correctly reported by a codec for
    display purposes.

    For if incorrect field ordering is specified, edges will appear ragged and
    not smooth.

    The sprite should be square with the white triangle having equal height and
    length and the hypotenuse lying at an angle of 45 degrees. The circular
    cut-out should be a perfect circle. This verifies that the pictures are
    displayed with the correct pixel aspect ratio (11.4.7).

    The text in the sprite is provided to check that the correct picture
    orientation has been used.

    The colors of the characters 'V', 'C' and '2' are colored saturated
    primary red, green, and blue for the color primaries used (11.4.10.2). This
    provides a basic verification that the color components have been provided
    to the display system and decoded correctly.
    """
    sprite = read_and_adapt_pointer_sprite(video_parameters)

    frame_width = video_parameters["frame_width"]
    frame_height = video_parameters["frame_height"]
    sprite_height, sprite_width, _ = sprite.shape

    # For bizarre, tiny picture formats too small for the sprite, clip the
    # sprite to fit
    sw = min(frame_width, sprite_width)
    sh = min(frame_height, sprite_height)

    # Blit the sprite onto an otherwise empty frame
    picture = np.zeros((frame_height, frame_width, 3))
    picture[:sh, :sw, :] = sprite[:sh, :sw, :]

    yield picture
    if video_parameters["source_sampling"] == SourceSamplingModes.interlaced:
        yield picture


def mid_gray(video_parameters, picture_coding_mode):
    """
    A video sequence containing exactly one empty mid-gray frame.

    'Mid gray' is defined as having each color component set to the integer
    value exactly half-way along its range. When encoded using VC-2 all
    transform coefficients will be zero.

    The actual color will differ depending on the color model used and the
    signal offsets specified, though typically a gray color is produced.
    """

    dd = compute_dimensions_and_depths(video_parameters, picture_coding_mode)

    y = np.full(
        (dd["Y"].height, dd["Y"].width),
        1 << (dd["Y"].depth_bits - 1),
    )
    c1 = np.full(
        (dd["C1"].height, dd["C1"].width),
        1 << (dd["C1"].depth_bits - 1),
    )
    c2 = np.full(
        (dd["C2"].height, dd["C2"].width),
        1 << (dd["C2"].depth_bits - 1),
    )

    yield {
        "Y": y.tolist(),
        "C1": c1.tolist(),
        "C2": c2.tolist(),
        "pic_num": 0,
    }
    if picture_coding_mode == PictureCodingModes.pictures_are_fields:
        yield {
            "Y": y.tolist(),
            "C1": c1.tolist(),
            "C2": c2.tolist(),
            "pic_num": 1,
        }


def white_noise(video_parameters, picture_coding_mode, num_frames=1, seed=0):
    """
    A video sequence containing uniformly distributed pseudo-random full-range
    signal values.

    .. note::

        Because all picture components are filled with noise, the resulting
        pictures will contain 'color' noise rather than black-and-white noise.
        This may include out-of-gamut signals.

    By default a single frame is produced, but the ``num_frames`` argument may
    be used to specify more.

    By default a fixed seed is used, but alternative seeds may be provided in
    the ``seed`` argument.
    """
    if picture_coding_mode == PictureCodingModes.pictures_are_fields:
        num_pictures = num_frames * 2
    else:
        num_pictures = num_frames

    dd = compute_dimensions_and_depths(video_parameters, picture_coding_mode)

    rand = np.random.RandomState(seed)

    for pic_num in range(num_pictures):
        picture = {"pic_num": pic_num}

        for component in ["Y", "C1", "C2"]:
            picture[component] = rand.randint(
                0,
                1 << dd[component].depth_bits,
                (dd[component].height, dd[component].width),
                # NB: int on Windows is only 32 bits, hence forced 64 bit width
                # here
                dtype=np.int64,
            ).tolist()

        yield picture


@pipe(xyz_to_native)
@pipe(progressive_to_pictures)
def linear_ramps(video_parameters, picture_coding_mode):
    """
    An video sequence containing exactly one frame with a series of linear
    color ramps (illustrated below).

    .. image:: /_static/test_images/linear_ramps.png

    The frame is split into horizontal bands which contain, top to bottom:

    * A black-to-white linear ramp
    * A black-to-red linear ramp
    * A black-to-green linear ramp
    * A black-to-blue linear ramp

    The color ramps are linear in intensity. For most color formats (using
    a non-linear transfer function) this produces a non-linear ramp in coded
    pixel values. Further, the ramp will not be perceptually linear since human
    vision does not have a linear response.

    This is provided for the purposes of checking that metadata related to
    color is correctly passed through for display purposes.
    """

    width = video_parameters["frame_width"]
    height = video_parameters["frame_height"]

    ramps_rgb = np.zeros((4, width, 3))

    ramps_rgb[0, :, :] = np.repeat(np.linspace(0.0, 1.0, width), 3).reshape(-1, 3)
    ramps_rgb[1, :, 0] = np.linspace(0.0, 1.0, width)
    ramps_rgb[2, :, 1] = np.linspace(0.0, 1.0, width)
    ramps_rgb[3, :, 2] = np.linspace(0.0, 1.0, width)

    ramps_xyz = matmul_colors(
        LINEAR_RGB_TO_XYZ[video_parameters["color_primaries_index"]],
        ramps_rgb,
    )

    frame_xyz = np.repeat(ramps_xyz, (height + 3) // 4, axis=0)[:height, :, :]

    yield frame_xyz
    if video_parameters["source_sampling"] == SourceSamplingModes.interlaced:
        yield frame_xyz



"""
Tests which verify that codecs used the necessary number of bits for their
arithmetic.
"""

import logging

from vc2_conformance.encoder import make_sequence

from vc2_conformance.bitstream import Stream

from vc2_conformance.picture_generators import (
    mid_gray,
    repeat_pictures,
)

from vc2_conformance.test_cases.bit_widths_common import (
    get_test_pictures,
    MissingStaticAnalysisError,
)

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.test_cases.decoder.common import iter_slices_in_sequence


@decoder_test_case_generator
def signal_range(codec_features):
    """
    **Tests that a decoder has sufficient numerical dynamic range.**

    These test cases contain a series of pictures containing test patterns
    designed to produce extreme signals within decoders. During these test
    cases, no integer clamping (except for final output clamping) or integer
    overflows must occur.

    A test case is produced for each picture component:

    ``signal_range[Y]``
        Luma component test patterns.

    ``signal_range[C1]``
        Color difference 1 component test patterns.

    ``signal_range[C2]``
        Color difference 2 component test patterns.

    These test cases are produced by encoding pictures consisting test patterns
    made up of entirely of legal (in range) signal values. Nevertheless, the
    resulting bitstreams produce large intermediate values within a decoder,
    though these are not guaranteed to be worst-case.

    .. note::

        For informational purposes, an example of a set of test patterns before
        and after encoding and quantisation is shown below:

        .. image:: /_static/user_guide/signal_range_decoder.svg

    .. note::

        The quantization indices used for lossy codecs are chosen to maximise
        the peak signal range produced by the test patterns. These are often
        higher than a typical VC-2 encoder might pick for a given bit rate but
        are nevertheless valid.

    An informative metadata file is provided along side each test case which
    gives, for each picture in the bitstream, the parts of a decoder which are
    being tested by the test patterns. See
    :py:class:`vc2_bit_widths.helpers.TestPoint` for details.
    """
    try:
        (
            analysis_luma_pictures,
            synthesis_luma_pictures,
            analysis_color_diff_pictures,
            synthesis_color_diff_pictures,
        ) = get_test_pictures(codec_features)
    except MissingStaticAnalysisError:
        logging.warning(
            (
                "No static analysis available for the wavelet "
                "used by codec '%s'. Signal range test cases cannot "
                "be generated."
            ),
            codec_features["name"],
        )
        return

    for component, analysis_test_pictures, synthesis_test_pictures in [
        ("Y", analysis_luma_pictures, synthesis_luma_pictures),
        ("C1", analysis_color_diff_pictures, synthesis_color_diff_pictures),
        ("C2", analysis_color_diff_pictures, synthesis_color_diff_pictures),
    ]:
        # For lossless codecs we use the analysis test patterns since no
        # quantisation takes place
        if codec_features["lossless"]:
            test_pictures = analysis_test_pictures
        else:
            test_pictures = synthesis_test_pictures

        # Generate an initially empty set of mid-gray pictures
        one_gray_frame = list(
            mid_gray(
                codec_features["video_parameters"],
                codec_features["picture_coding_mode"],
            )
        )
        pictures = list(
            repeat_pictures(
                one_gray_frame,
                ((len(test_pictures) + len(one_gray_frame) - 1) // len(one_gray_frame)),
            )
        )

        # Fill-in the test patterns
        minimum_qindices = []
        for test_picture, picture in zip(test_pictures, pictures):
            picture[component] = test_picture.picture.tolist()
            if codec_features["lossless"]:
                minimum_qindices.append(0)
            else:
                minimum_qindices.append(test_picture.quantisation_index)
        while len(minimum_qindices) < len(pictures):
            minimum_qindices.append(0)

        # Extract the testpoints in JSON-serialisable form
        metadata = [[tp._asdict() for tp in p.test_points] for p in test_pictures]

        # Encode
        sequence = make_sequence(
            codec_features,
            pictures,
            minimum_qindex=minimum_qindices,
        )

        # Check the desired qindex could be used (should only ever fail for
        # absurdly low bitrate configurations).
        num_unexpected_qindices = 0
        expected_qindex = None
        expected_qindex_iter = iter(minimum_qindices)
        for _, sx, sy, slice in iter_slices_in_sequence(codec_features, sequence):
            if sx == 0 and sy == 0:
                expected_qindex = next(expected_qindex_iter, 0)
            if slice["qindex"] != expected_qindex:
                num_unexpected_qindices += 1

        if num_unexpected_qindices > 0:
            logging.warning(
                "Could not assign the required qindex to %d picture slices "
                "for signal range test case due to a small picture_bytes value. "
                "Peak signal levels might be reduced.",
                num_unexpected_qindices,
            )

        yield TestCase(
            Stream(sequences=[sequence]),
            component,
            metadata=metadata,
        )



"""
Tests which verify sequence header contents (11) are correctly read.
"""

from copy import deepcopy

from vc2_data_tables import ParseCodes

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import (
    TestCase,
    decoder_test_case_generator,
)

from vc2_conformance.picture_generators import (
    static_sprite,
    repeat_pictures,
)

from vc2_conformance.encoder import (
    IncompatibleLevelAndDataUnitError,
    iter_sequence_headers,
    make_sequence,
)


def replace_sequence_headers(sequence, sequence_header):
    r"""
    Replace the :py:class:`SequenceHeaders
    <vc2_conformance.bitstream.SequenceHeader>` in a
    :py:class:`~vc2_conformance.bitstream.Sequence` with the provided
    alternative.

    A new sequence is returned and the old sequence left unmodified. Likewise
    the ``sequence_header`` argument will be copied.
    """
    sequence = deepcopy(sequence)
    for data_unit in sequence["data_units"]:
        if data_unit["parse_info"]["parse_code"] == ParseCodes.sequence_header:
            data_unit["sequence_header"] = deepcopy(sequence_header)

    return sequence


@decoder_test_case_generator
def source_parameters_encodings(codec_features):
    """
    **Tests the decoder can decode different encodings of the video format
    metadata.**

    This series of test cases each contain the same source parameters (11.4),
    but in different ways.

    ``source_parameters_encodings[custom_flags_combination_?_base_video_format_?]``
        For these test cases, the base video format which most closely matches
        the desired video format is used. Each test case incrementally checks
        that source parameters can be explicitly set to their desired values
        (e.g. by setting ``custom_*_flag`` bits to 1).

    ``source_parameters_encodings[base_video_format_?]``
        These test cases, check that other base video formats can be used (and
        overridden) to specify the desired video format. Each of these test
        cases will explicitly specify as few video parameters as possible (e.g.
        setting as many ``custom_*_flag`` fields to 0 as possible).

    .. tip::

        The :ref:`vc2-bitstream-viewer` can be used to display the encoding
        used in a given test case as follows::

            $ vc2-bitstream-viewer --show sequence_header path/to/test_case.vc2

    .. note::

        Some VC-2 levels constrain the allowed encoding of source parameters in
        the bit stream and so fewer test cases will be produced in this
        instance.

    .. note::

        Not all base video formats can be used as the basis for encoding a
        specific video format. For example, the 'top field first' flag (11.3)
        set by a base video format cannot be overridden. As a result, test
        cases will not include every base video format index.

    """
    # Generate a base sequence in which we'll replace the sequence headers
    # later
    base_sequence = make_sequence(
        codec_features,
        static_sprite(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
    )

    # To keep the number of tests sensible, we'll include all sequence header
    # encodings using the best-matching base video format followed by the
    # least-custom-overridden encoding for all other base video formats. This
    # checks out as many 'custom' flags as possible (against the best-matching
    # base video format) and also checks (as best possible) the other base
    # video format values are correct.
    best_base_video_format = None
    last_base_video_format = None
    for i, sequence_header in enumerate(iter_sequence_headers(codec_features)):
        base_video_format = sequence_header["base_video_format"]

        # The iter_sequence_headers function returns headers with the best
        # matching base video format first
        if best_base_video_format is None:
            best_base_video_format = base_video_format

        # The iter_source_parameter_options produces sequence headers with
        # base video formats grouped consecutively. The first example of each
        # will use the fewest possible 'custom' flags and therefore best tests
        # that the base video format parameters are correct in the decoder.
        first_example_of_base_video_format = base_video_format != last_base_video_format
        last_base_video_format = base_video_format

        if base_video_format == best_base_video_format:
            yield TestCase(
                Stream(
                    sequences=[replace_sequence_headers(base_sequence, sequence_header)]
                ),
                "custom_flags_combination_{}_base_video_format_{:d}".format(
                    i + 1,
                    base_video_format,
                ),
            )
        elif first_example_of_base_video_format:
            yield TestCase(
                Stream(
                    sequences=[replace_sequence_headers(base_sequence, sequence_header)]
                ),
                "base_video_format_{:d}".format(base_video_format),
            )


@decoder_test_case_generator
def repeated_sequence_headers(codec_features):
    """
    **Tests the decoder can handle a stream with repeated sequence headers.**

    This test case consists of a sequence containing two frames in which the
    sequence header is repeated before every picture.

    This test will be omitted if the VC-2 level prohibits the repetition of the
    sequence header.
    """
    try:
        # Generate a base sequence in which we'll replace the sequence headers
        # later. We ensure we have at least two pictures to ensure we get
        # pictures and sequence headers being interleaved.
        sequence = make_sequence(
            codec_features,
            repeat_pictures(
                static_sprite(
                    codec_features["video_parameters"],
                    codec_features["picture_coding_mode"],
                ),
                2,
            ),
            # Force an extra sequence header between every data unit
            "(sequence_header .)+",
        )
    except IncompatibleLevelAndDataUnitError:
        # Do not try to force levels which don't support this level of sequence
        # header interleaving to accept it.
        return None

    return Stream(sequences=[sequence])



from itertools import count, repeat, cycle

from vc2_data_tables import QUANTISATION_MATRICES

from vc2_conformance.codec_features import CodecFeatures

from vc2_conformance.test_cases import TestCase, decoder_test_case_generator

from vc2_conformance.constraint_table import allowed_values_for, ValueSet, AnyValue

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.codec_features import codec_features_to_trivial_level_constraints

from vc2_conformance.test_cases.decoder.pictures import static_noise

from vc2_conformance.test_cases.decoder.lossless_quantization import (
    lossless_quantization,
)

from vc2_conformance.test_cases import normalise_test_case_generator


def generate_test_stream(codec_features):
    """
    Generate an appropriate test sequence with the specified codec features
    (i.e. quantisation matrix).

    For lossy coding modes, a noise plate will be used. For lossless modes, the
    lossless_quantization test stream will be used instead.
    """
    if codec_features["lossless"]:
        # The special lossless quantization test signal is used for lossless
        # modes since using white noise could result in transform values
        # outside the legal range. (Lossy quantization can result in signal
        # levels several orders of magnitude larger than a lossless encoding).
        f = lossless_quantization
    else:
        # White noise is used for lossy coding modes as it trivially results in
        # signal in every transform band.
        f = static_noise

    test_cases = list(normalise_test_case_generator(f, codec_features))
    if len(test_cases) == 1:
        return test_cases[0].value
    else:
        # Only occurs if the lossless_quantization generator fails to produce a
        # test case. Nothing to be done, we just have to abandon the test case.
        return None


@decoder_test_case_generator
def default_quantization_matrix(codec_features):
    """
    **Tests that the default quantization matrix can be used.**

    This test case is only generated when a non ``default`` value is specified
    for the ``quantization_matrix`` codec features CSV entry but when a default
    quantization matrix is defined.

    .. note::

        This is the only test case which sets the ``custom_quant_matrix`` flag
        (12.4.5.3) to 0 when a ``quantization_matrix`` is supplied in the codec
        features CSV.

    .. note::

        For lossy coding modes, the encoded picture will contain a noise signal
        (see the :decoder-test-case:`static_noise` test case).

        For lossless coding modes, the encoded picture will be the test pattern
        used by the :decoder-test-case:`lossless_quantization` test case. This
        test pattern is designed to be losslessly encodable when some
        quantization is applied.
    """
    # Skip if already using the default quantisation matrix
    if codec_features["quantization_matrix"] is None:
        return None

    # Skip if no default quantization matrix is defined for the codec in use
    if (
        codec_features["wavelet_index"],
        codec_features["wavelet_index_ho"],
        codec_features["dwt_depth"],
        codec_features["dwt_depth_ho"],
    ) not in QUANTISATION_MATRICES:
        return None

    # Skip if the level requires a custom quantisation matrix to be specified
    constrained_values = codec_features_to_trivial_level_constraints(codec_features)
    del constrained_values["custom_quant_matrix"]
    if False not in allowed_values_for(
        LEVEL_CONSTRAINTS, "custom_quant_matrix", constrained_values
    ):
        return None

    # Override the quantization_matrix setting in the codec features to force
    # the default quantization matrix to be used
    codec_features = CodecFeatures(
        codec_features,
        quantization_matrix=None,
    )

    return generate_test_stream(codec_features)


def quantization_matrix_from_generator(codec_features, generator):
    """
    Generate a quantization matrix for the specified set of codec_features
    where the values are taken from the provided generator function.
    """
    values = iter(generator)

    dwt_depth = codec_features["dwt_depth"]
    dwt_depth_ho = codec_features["dwt_depth_ho"]

    out = {}
    if dwt_depth_ho == 0:
        out[0] = {"LL": int(next(values))}
    else:
        out[0] = {"L": int(next(values))}
        for level in range(1, dwt_depth_ho + 1):
            out[level] = {"H": int(next(values))}
    for level in range(dwt_depth_ho + 1, dwt_depth + dwt_depth_ho + 1):
        out[level] = {
            "HL": int(next(values)),
            "LH": int(next(values)),
            "HH": int(next(values)),
        }

    return out


@decoder_test_case_generator
def custom_quantization_matrix(codec_features):
    """
    **Tests that a custom quantization matrix can be specified.**

    A series of bitstreams with different custom quantisation matrices are
    generated as follows:

    ``custom_quantization_matrix[zeros]``
        Specifies a custom quantisation matrix with all matrix values set to zero.

    ``custom_quantization_matrix[arbitrary]``
        Specifies a custom quantisation matrix with all matrix values set to
        different, though arbitrary, values.

    ``custom_quantization_matrix[default]``
        Specifies a custom quantisation matrix containing the same values as
        the default quantisation matrix. This test case is only generated when
        a default quantization matrix is defined for the codec.

    These test cases are only generated when permitted by the VC-2 level in
    use.

    .. note::

        For lossy coding modes, the encoded picture will contain a noise signal
        (see the :decoder-test-case:`static_noise` test case).

        For lossless coding modes, the encoded picture will be the test pattern
        used by the :decoder-test-case:`lossless_quantization` test case. This
        test pattern is designed to be losslessly encodable when some
        quantization is applied.
    """
    # Skip if the level disallows custom quantisation matrices
    constrained_values = codec_features_to_trivial_level_constraints(codec_features)
    constrained_values["custom_quant_matrix"] = True
    allowed_quant_matrix_values = allowed_values_for(
        LEVEL_CONSTRAINTS, "quant_matrix_values", constrained_values
    )
    if allowed_quant_matrix_values == ValueSet():
        return

    # A set of alternative quantisation matrices to test
    # [(description, quantization_matrix), ...]
    candidate_matrices = []

    # A special case: a flat (all zeros) quantisation matrix
    candidate_matrices.append(
        ("zeros", quantization_matrix_from_generator(codec_features, repeat(0)))
    )

    # An arbitrary quantisation matrix with different values in each entry (if
    # possible)
    if allowed_quant_matrix_values == AnyValue():
        values = count()
    else:
        values = cycle(sorted(allowed_quant_matrix_values.iter_values()))
    candidate_matrices.append(
        ("arbitrary", quantization_matrix_from_generator(codec_features, values))
    )

    # If a default quantisation matrix is available, try explicitly giving the
    # values from the default quant matrix
    wavelet_specification = (
        codec_features["wavelet_index"],
        codec_features["wavelet_index_ho"],
        codec_features["dwt_depth"],
        codec_features["dwt_depth_ho"],
    )
    if wavelet_specification in QUANTISATION_MATRICES:
        candidate_matrices.append(
            (
                "default",
                QUANTISATION_MATRICES[wavelet_specification],
            )
        )

    # Generate test cases for each custom quantisation matrix.
    for description, quantization_matrix in candidate_matrices:
        # If we happen to have generated the quantization matrix specified in
        # the codec features, skip that case...
        if quantization_matrix == codec_features["quantization_matrix"]:
            continue

        # If we the level restricts the quantisation matrix values such that
        # the generated cases is not allowed, skip it.
        if not all(
            value in allowed_quant_matrix_values
            for orients in quantization_matrix.values()
            for value in orients.values()
        ):
            continue

        stream = generate_test_stream(
            CodecFeatures(
                codec_features,
                quantization_matrix=quantization_matrix,
            )
        )
        if stream is not None:
            yield TestCase(stream, description)
